from datetime import datetime
from cnvrgv2.config import routes, error_messages
from cnvrgv2.config.error_messages import NOT_A_DATASET_OBJECT
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes
from cnvrgv2.modules.dataset import Dataset
from cnvrgv2.modules.flows.flow_version.flow_versions_client import FlowVersionsClient
from cnvrgv2.proxy import Proxy, HTTP
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.utils.cron import Cron
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.url_utils import urljoin
from cnvrgv2.utils.validators import validate_types_in_list


class Flow(DynamicAttributes):
    available_attributes = {
        "title": str,
        "slug": str,
        "created_at": datetime,
        "updated_at": datetime,
        "cron_syntax": str,
        "webhook_url": str,
        "trigger_dataset": str
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        if slug:
            self._context.set_scope(SCOPE.FLOW, slug)

        scope = self._context.get_scope(SCOPE.FLOW)

        self._proxy = Proxy(context=self._context)
        self._route = routes.FLOW_BASE.format(scope["organization"], scope["project"], scope["flow"])
        self._attributes = attributes or {}
        self._type = "Flow"
        self.slug = scope["flow"]

        self._init_clients()

    def run(self):
        """
        Runs the current flow
        @return: None
        """

        run_url = urljoin(self._route, routes.RUN_FLOW)
        self._proxy.call_api(
            route=run_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes={})
        )

    def update(self, **kwargs):
        """
        Updates the current flows
        @param kwargs: Dictionary of attributes to update.
        supported attributes: title
        @return: The updated flows
        """

        if not kwargs:
            raise CnvrgArgumentsError(error_messages.EMPTY_ARGUMENT.format("kwargs"))

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.PUT,
            payload=JAF.serialize(type=self._type, attributes=kwargs)
        )

        self._attributes = response.attributes
        return self

    def set_schedule(self, cron_expression):
        """
        Schedule the latest version of the current flows to run on a recurring basis
        @param cron_expression: A string in cron format. You can use Cron helper class
        @return:None
        """
        Cron.validate_cron_syntax(cron_expression)
        set_schedule_url = urljoin(self._route, routes.FLOW_SET_SCHEDULE)
        attributes = {
            "cron_syntax": cron_expression
        }

        self._proxy.call_api(
            route=set_schedule_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def clear_schedule(self):
        """
        Clears the recurring schedule, if exists
        @return:None
        """
        set_schedule_url = urljoin(self._route, routes.FLOW_STOP_SCHEDULE)

        self._proxy.call_api(
            route=set_schedule_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes={})
        )

    def toggle_webhook(self, toggle):
        """
        Toggles webhook on/off
        @param toggle: Boolean. Turn webhook on or off
        @return: None
        """
        if not isinstance(toggle, bool):
            error_message = error_messages.ARGUMENT_BAD_TYPE.format(bool.__name__, type(toggle).__name__)
            raise CnvrgArgumentsError(error_message)

        toggle_webhook_url = urljoin(self._route, routes.FLOW_TOGGLE_WEBHOOK)
        attributes = {
            "toggle": toggle
        }

        self._proxy.call_api(
            route=toggle_webhook_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def toggle_dataset_update(self, toggle, dataset=None):
        """
        Toggle a trigger that will run flow upon a dataset update
        @param toggle: Turn flow trigger upon dataset update
        @param dataset: The dataset that will trigger the flow upon update (dataset object)
        @return: None
        """
        argument_errors = {}
        if not isinstance(toggle, bool):
            error_message = error_messages.ARGUMENT_BAD_TYPE.format(bool.__name__, type(toggle).__name__)
            argument_errors["toggle"] = error_message

        if toggle and not (dataset and validate_types_in_list([dataset], Dataset)):
            argument_errors["dataset"] = NOT_A_DATASET_OBJECT

        if argument_errors:
            raise CnvrgArgumentsError(argument_errors)

        toggle_dataset_update_url = urljoin(self._route, routes.FLOW_TOGGLE_DATASET_UPDATE)
        attributes = {}

        if toggle:
            attributes = {
                "dataset_slug": dataset.slug
            }

        self._proxy.call_api(
            route=toggle_dataset_update_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def delete(self):
        """
        Deletes the current flow
        @return: None
        """
        self._proxy.call_api(
           route=self._route,
           http_method=HTTP.DELETE
        )

    def _init_clients(self):
        self.flow_versions = FlowVersionsClient(self._context)



import pytest

from cnvrgv2.proxy import Proxy
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.project import Project
from cnvrgv2.modules.projects_client import ProjectsClient
from cnvrgv2.utils.json_api_format import JAF


class TestProjectsClient:
    def test_init_no_credentials(self):
        with pytest.raises(CnvrgArgumentsError):
            ProjectsClient(None)

    def test_init_valid_credentials(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = ProjectsClient(empty_cnvrg)
        assert client._context
        assert client._route

    def test_get_with_slug(self, mocker, empty_context, empty_cnvrg):
        slug = "test-slug"

        def mock_call_api(*args, **kwargs):
            response = {
                "data": {
                    "id": "17",
                    "type": "project",
                    "attributes": {
                        "title": "test-title",
                        "slug": slug,
                        "git": False,
                        "start_commit": "b6692ea5df920cad691c20319a6fffd7a4a766b8",
                        "commit": "5b384ce32d8cdef02bc3a139d4cac0a22bb029e8"
                    }
                }
            }
            return JAF(response=response)

        mocked_api_call = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        empty_cnvrg._context = empty_context
        client = ProjectsClient(empty_cnvrg)
        model = client.get(slug)

        # The function will not be called on init since its lazy-loading
        assert mocked_api_call.call_count == 0

        # The function will not be called on init since its lazy-loading
        assert model.title == "test-title"
        assert mocked_api_call.call_count > 0

        assert type(model) is Project

    def test_get_with_empty_slug(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = ProjectsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.get("")

    def test_get_with_none_slug(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = ProjectsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.get(None)

    def test_get_with_wrong_type_slug(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = ProjectsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.get({"say": "whaaat??"})

    def test_valid_create(self, mocker, empty_context, empty_cnvrg):
        def mock_call_api(*args, **kwargs):
            response = {
                "data": {
                    "id": "17",
                    "type": "project",
                    "attributes": {
                        "title": "test-title",
                        "slug": "test-slug",
                        "git": False,
                        "start_commit": "b6692ea5df920cad691c20319a6fffd7a4a766b8",
                        "commit": "5b384ce32d8cdef02bc3a139d4cac0a22bb029e8"
                    }
                }
            }
            return JAF(response=response)

        mocked_api_call = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        empty_cnvrg._context = empty_context
        client = ProjectsClient(empty_cnvrg)
        model = client.create("test-title")

        # TODO: add attribute checks. added call_count until decided what to check
        assert type(model) is Project
        assert mocked_api_call.call_count > 0

    def test_create_with_empty_name(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = ProjectsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create("")

    def test_create_with_none_name(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = ProjectsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create(None)

    def test_create_with_wrong_type_name(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = ProjectsClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create({"say": "whaaat??"})



import pytest

from cnvrgv2.proxy import Proxy
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.queries.query import Query
from cnvrgv2.modules.queries.queries_client import QueriesClient
from cnvrgv2.utils.json_api_format import JAF


class TestQueriesClient:
    def test_init_valid_credentials(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = QueriesClient(empty_cnvrg)
        assert client._context
        assert client._route

    def test_get_with_slug(self, mocker, empty_context, empty_cnvrg):
        slug = "test-slug"

        def mock_call_api(*args, **kwargs):
            response = {
                "data": {
                    "id": "17",
                    "type": "query",
                    "attributes": {
                        "slug": slug,
                        "name": "test-name",
                        "query": "test-query",
                        "commit_sha1": "5b384ce32d8cdef02bc3a139d4cac0a22bb029e8",
                    }
                }
            }
            return JAF(response=response)

        mocked_api_call = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        empty_cnvrg._context = empty_context
        client = QueriesClient(empty_cnvrg)
        model = client.get(slug)

        # The function will not be called on init since its lazy-loading
        assert mocked_api_call.call_count == 0

        # The function will be called when accessing an attribute
        assert model.name == "test-name"
        assert mocked_api_call.call_count > 0

        assert type(model) is Query

    def test_get_with_empty_slug(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = QueriesClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.get("")

    def test_get_with_none_slug(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = QueriesClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.get(None)

    def test_get_with_wrong_type_slug(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = QueriesClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.get({"say": "whaaat??"})

    def test_valid_create(self, mocker, empty_context, empty_cnvrg):
        name = "test-name"
        query = "test-query"

        def mock_call_api(*args, **kwargs):
            response = {
                "data": {
                    "id": "17",
                    "type": "project",
                    "attributes": {
                        "slug": "test-slug",
                        "name": name,
                        "query": query,
                        "commit_sha1": "5b384ce32d8cdef02bc3a139d4cac0a22bb029e8",
                    }
                }
            }
            return JAF(response=response)

        mocked_api_call = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        empty_cnvrg._context = empty_context
        client = QueriesClient(empty_cnvrg)
        model = client.create(name=name, query=query, commit_sha1=None)

        # TODO: add attribute checks. added call_count until decided what to check
        assert type(model) is Query
        assert mocked_api_call.call_count > 0

    def test_create_with_empty_name(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = QueriesClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create(name="", query="test-query")

    def test_create_with_none_name(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = QueriesClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create(name=None, query="test-query")

    def test_create_with_empty_query(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = QueriesClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create(name="test-name", query="")

    def test_create_with_none_query(self, empty_context, empty_cnvrg):
        empty_cnvrg._context = empty_context
        client = QueriesClient(empty_cnvrg)
        with pytest.raises(CnvrgArgumentsError):
            client.create(name="test-name", query=None)



import json
from typing import Optional
from cnvrgv2 import cnvrg
from cnvrgv2.config import error_messages, routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.datasource.datasource import Datasource, StorageTypes
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.json_api_format import JAF


class DatasourcesClient:
    def __init__(self, organization):
        if type(organization) is not cnvrg.Cnvrg:
            raise CnvrgArgumentsError(error_messages.ARGUMENT_BAD_TYPE.format(cnvrg.Cnvrg, type(organization)))
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.DATASOURCES_BASE.format(scope["organization"])

    def list(self, filter=None, sort="-id"):
        """
        List all datasources in a specific organization, assigned to the current user
        @param filter: Filter json
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields data source objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=Datasource,
            sort=sort,
            filter=filter
        )

    def list_count(self, filter=None, sort="-id"):
        """
        Count of datasources in a specific organization, assigned to the current user
        @param filter: Filter json
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Total count of data sources
        """
        proxy = Proxy(self._context)
        data = {}
        page_size = 20
        filters = json.dumps(filter) if filter else None
        payload = {"sort": sort, "page[size]": page_size, "filter": filters, **data}
        response = proxy.call_api(
            route=self._route,
            http_method=HTTP.GET,
            payload=payload
        )
        return response.meta['total']

    def get(self, slug):
        """
        Retrieves a data source by the given slug
        @param slug: The slug of the requested data source
        @return: Datasource object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.DATASOURCE_GET_FAULTY_SLUG)

        return Datasource(context=self._context, slug=slug)

    def create(self,
               name: str,
               storage_type: str,
               bucket_name: str,
               credentials: dict,
               region: str,
               description: str = None,
               public: bool = None,
               path: str = None,
               endpoint: str = None,
               collaborators: Optional[list] = None,
               access_point: str = None):
        """
        Creates a new datasource with the given name
        @param storage_type: Enum storage type of the new data source (S3, minio etc), StorageTypes
        @param name: name of the new data source
        @param bucket_name: bucket_name of the new data source
        @param path: path of the new data source
        @param endpoint: endpoint of the new data source
        @param region: region of the new data source
        @param description: description of the new data source
        @param public: permission level of the new data source
        @param credentials: credentials of the new data source
        @param collaborators: users of the new data source (collaborators), array of emails
        @param access_point: AWS Access-point
        @return: the newly created Datasource object
        """
        # mandatory fields
        if not name or not isinstance(name, str):
            raise CnvrgArgumentsError(error_messages.DATASOURCE_NOT_VALID_PARAM.format("name", "string"))
        if not bucket_name or not isinstance(bucket_name, str):
            raise CnvrgArgumentsError(error_messages.DATASOURCE_NOT_VALID_PARAM.format("bucket_name", "string"))
        if not credentials or not isinstance(credentials, dict):
            raise CnvrgArgumentsError(error_messages.DATASOURCE_NOT_VALID_PARAM.format("credentials", "dict"))
        if not region or not isinstance(region, str):
            raise CnvrgArgumentsError(error_messages.DATASOURCE_NOT_VALID_PARAM.format("region", "string"))
        # optional fields
        if path and not isinstance(path, str):
            raise CnvrgArgumentsError(error_messages.DATASOURCE_NOT_VALID_PARAM.format("path", "string"))
        if endpoint and not isinstance(endpoint, str):
            raise CnvrgArgumentsError(error_messages.DATASOURCE_NOT_VALID_PARAM.format("endpoint", "string"))
        if description and not isinstance(description, str):
            raise CnvrgArgumentsError(error_messages.DATASOURCE_NOT_VALID_PARAM.format("description", "string"))
        if public and not isinstance(public, bool):
            raise CnvrgArgumentsError(error_messages.DATASOURCE_NOT_VALID_PARAM.format("public", "boolean"))
        if collaborators and not isinstance(collaborators, list):
            raise CnvrgArgumentsError(error_messages.DATASOURCE_NOT_VALID_PARAM.format("users", "array"))
        if not storage_type or not isinstance(storage_type, StorageTypes):
            arr = list(StorageTypes.__members__.keys())
            valid_types = ', '.join(arr)
            raise CnvrgArgumentsError(error_messages.DATASOURCE_FAULTY_STORAGE_TYPE.format(valid_types))

        attributes = {
            "bucket_name": bucket_name,
            "name": name,
            "storage_type": storage_type.value if storage_type else None,
            "path": path,
            "endpoint": endpoint,
            "region": region,
            "description": description,
            "credentials": credentials,
            "users_emails": collaborators,
            "access_point": access_point
        }

        if public is not None:
            attributes['public'] = public

        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="datasource", attributes=attributes)
        )

        slug = response.attributes['slug']
        return Datasource(context=self._context, slug=slug)



import os
import shutil
from os import path

import pytest
from tests.conftest import random_string
from cnvrgv2 import Cnvrg
from cnvrgv2.errors import CnvrgHttpError
from cnvrgv2.modules.project import Project
from tests.e2e.conftest import call_database


class TestProjects:

    @staticmethod
    def cleanup(context_prefix):
        delete_user_command = "DELETE FROM projects WHERE slug LIKE'{}%'".format(context_prefix)
        call_database(delete_user_command)

    @staticmethod
    def create_project(class_context, e2e_client):
        name = class_context.generate_name(5)
        return e2e_client.projects.create(name=name)

    def test_create(self, class_context, e2e_client):
        model = TestProjects.create_project(class_context, e2e_client)
        assert type(model) == Project

    def test_get(self, class_context, e2e_client):
        project = TestProjects.create_project(class_context, e2e_client)

        model = e2e_client.projects.get(slug=project.slug)
        assert type(model) == Project
        assert model.slug == project.slug
        assert model.title == project.slug

    def test_get_non_existent(self, random_name, e2e_client):
        name = random_name(5)

        model = e2e_client.projects.get(slug=name)
        assert type(model) == Project

        with pytest.raises(CnvrgHttpError) as exception_info:
            model.title

        assert "found" in str(exception_info.value)

    def test_delete(self, class_context, e2e_client):
        model = TestProjects.create_project(class_context, e2e_client)
        name = model.slug
        model.delete()

        model = e2e_client.projects.get(slug=name)
        with pytest.raises(CnvrgHttpError) as exception_info:
            model.title

        assert "found" in str(exception_info.value)

    def test_delete_non_existent(self, random_name, e2e_client):
        name = random_name(5)

        model = e2e_client.projects.get(slug=name)
        with pytest.raises(CnvrgHttpError) as exception_info:
            model.delete()

        assert "found" in str(exception_info.value)

    def test_delete_not_authorized(self, domain, e2e_data_scientist_user, e2e_project):
        cnvrg = Cnvrg(
            domain=domain,
            email=e2e_data_scientist_user["email"],
            password=e2e_data_scientist_user["password"],
        )

        with pytest.raises(CnvrgHttpError) as exception_info:
            cnvrg.projects.delete([e2e_project.slug])

        assert "unauthorized" in str(exception_info.value).lower()

    def test_empty_bulk_delete_fail(self, e2e_client):

        with pytest.raises(CnvrgHttpError) as exception_info:
            e2e_client.projects.delete([])

        assert "no project slug was sent." in str(exception_info.value).lower()

    def test_bulk_delete_projects_success(self, class_context, e2e_client):
        num_of_projects = 5
        project_slugs = []

        # Create some projects
        for i in range(num_of_projects):
            project = TestProjects.create_project(class_context, e2e_client)
            project_slugs.append(project.slug)

        e2e_client.projects.delete(project_slugs)

        for slug in project_slugs:
            with pytest.raises(CnvrgHttpError) as exception_info:
                e2e_client.projects.get(slug).title
            assert exception_info.value.status_code == 404

    def test_clone_success(self, random_name, e2e_project):
        filename = expected_output_dir = ""
        try:
            filename = random_name(5)
            expected_output_dir = path.join(e2e_project.title)
            expected_output_file = path.join(expected_output_dir, filename)
            open(filename, 'w')
            e2e_project.put_files([filename], message="test project files")

            e2e_project.clone()

            assert path.exists(expected_output_file)

        finally:
            # cleanup
            if path.exists(filename):
                os.remove(filename)
            if path.exists(expected_output_dir):
                shutil.rmtree(expected_output_dir)

    def test_list(self, class_context, e2e_client):
        # Clean the db before testing list
        TestProjects.cleanup(class_context.prefix)
        name_base = class_context.generate_name(5)
        for i in range(40):
            name = "{}{}".format(name_base, i)
            e2e_client.projects.create(name=name)
        # Check descending order works
        idx = 39
        projects = e2e_client.projects.list()
        for project in projects:
            assert project.slug == "{}{}".format(name_base, idx)
            idx -= 1
        # Check ascending order works
        idx = 0
        projects = e2e_client.projects.list(sort="id")
        for project in projects:
            assert project.slug == "{}{}".format(name_base, idx)
            idx += 1

    def test_sync_should_download(self, e2e_project, e2e_git_project):
        """
        this test simulates the following scenario:
        1. on running experiment when we do sync, download should not happen
        2. on running workspace with git project when we do sync, download should not happen
        """
        folders_to_delete = []
        old_working_dir = os.getcwd()
        try:
            # Create scenarios of envs
            jobs = ["NotebookSession", "Experiment"]
            for job in jobs:
                # Setup project and project files
                project = None
                if job == "NotebookSession":
                    # should be git project for should_download to be false
                    project = e2e_git_project
                elif job == "Experiment":
                    project = e2e_project
                assert project is not None

                # create experiment temp folder
                exp_folder_name = random_string(5)
                folders_to_delete.append(exp_folder_name)
                if not os.path.isdir(exp_folder_name):
                    os.mkdir(exp_folder_name)
                os.chdir(exp_folder_name)
                project.clone()

                os.chdir(old_working_dir)

                # Create new project latest commit
                folder_name = random_string(5)
                folders_to_delete.append(folder_name)
                if not os.path.isdir(folder_name):
                    os.mkdir(folder_name)
                os.chdir(folder_name)
                project.clone()
                os.chdir(project.slug)

                # upload new file as master commit
                master_filename = random_string(5)
                open(master_filename, "a")
                project.put_files(paths=[path.join(master_filename)])

                # Simulate sync inside the experiment
                os.chdir(path.join(old_working_dir, exp_folder_name, project.slug))
                # Set env for test
                os.environ["CNVRG_JOB_TYPE"] = job

                # Test sync, should not download master file, only upload.
                project.sync()
                assert os.path.exists(master_filename) is False

                del os.environ["CNVRG_JOB_TYPE"]
                os.chdir(old_working_dir)
        finally:
            # Clean up
            if os.environ.get("CNVRG_JOB_TYPE"):
                del os.environ["CNVRG_JOB_TYPE"]

            os.chdir(old_working_dir)
            for folder in folders_to_delete:
                if os.path.isdir(folder):
                    shutil.rmtree(folder)



import os

import pytest
import requests

from cnvrgv2.config import error_messages
from cnvrgv2.config.config import SSO_VERSION
from cnvrgv2.context import Context
from cnvrgv2.errors import CnvrgError, CnvrgLoginError
from cnvrgv2.modules.users.users_client import UsersClient


class MockResponse:
    def __init__(self, status, json):
        self.status_code = status
        self._json = json

    def json(self, *args, **kwargs):
        return self._json


class FakeContext:
    def __init__(self):
        # Set the credentials variables:
        self.token = "fake"
        self.domain = "fake"
        self.user = "fake"

        # Set context variables:
        self.project = "fake"
        self.experiment = "fake"
        self.organization = "fake"
        self.sso_version = None


# TODO: add organization auth test
class TestContext:
    def test_init(self, mocker):
        # Test that the constructor raises errors on bad params:
        with pytest.raises(CnvrgError):
            Context(domain="test", user="test")
        with pytest.raises(CnvrgError):
            Context(user="test", password="test", organization="test")

        # Test that the load_credentials function was called on empty input
        def mock_load_credentials(obj):
            obj.domain = "test-success"

        def mock_organization_exists(obj, name):
            return True

        mocker.patch.object(Context, "_load_credentials", mock_load_credentials)
        mocker.patch.object(Context, "ensure_organization_exist", mock_organization_exists)
        context = Context()
        assert context.domain == "test-success"

    def test_init_from_context(self):
        fake_context = FakeContext()
        copied_context = Context(context=fake_context)

        assert copied_context.token == fake_context.token
        assert copied_context.domain == fake_context.domain
        assert copied_context.user == fake_context.user

        assert copied_context.project == fake_context.project
        assert copied_context.experiment == fake_context.experiment
        assert copied_context.organization == fake_context.organization

        copied_context.token = "new-token"
        assert copied_context.token != fake_context.token

    def test_authenticate_success(self, mocker, empty_context):
        # Test that the load_credentials function runs correctly with valid response
        def mock_requests_post(*args, **kwargs):
            resp = {
                "meta": {
                    "jwt": "fake-token",
                    "organization": "org1"
                }
            }
            return MockResponse(status=200, json=resp)

        def mock_get_sso_version(*args, **kwargs):
            return SSO_VERSION.CAPI

        mocker.patch.object(requests, "post", mock_requests_post)
        mocker.patch.object(UsersClient, "_get_sso_version", side_effect=mock_get_sso_version)

        # Test call
        empty_context._authenticate("http://fake-url.com", "fake-user", "fake-pass")
        assert empty_context.token == "fake-token"
        assert empty_context.organization == "org1"

    def test_authenticate_fail(self, mocker, empty_context):
        # Test that the load_credentials function runs correctly with invalid response
        def mock_requests_post(*args, **kwargs):
            resp = {
                "message": "An exception has occurred"
            }
            return MockResponse(status=401, json=resp)

        def mock_get_sso_version(*args, **kwargs):
            return SSO_VERSION.CAPI

        mocker.patch.object(requests, "post", mock_requests_post)
        mocker.patch.object(UsersClient, "_get_sso_version", side_effect=mock_get_sso_version)

        # Test call
        with pytest.raises(CnvrgLoginError) as exception_info:
            empty_context._authenticate("http://fake-url.com", "fake-user", "fake-pass")
        assert error_messages.INVALID_CREDENTIALS in str(exception_info.value)

    def test_load_from_env(self, monkeypatch, empty_context):
        monkeypatch.setenv("CNVRG_URL", "fake-domain")
        monkeypatch.setenv("CNVRG_JWT_TOKEN", '')
        monkeypatch.setenv("CNVRG_ORGANIZATION", '')

        assert not empty_context._load_from_env()

        monkeypatch.setenv("CNVRG_URL", "fake-domain")
        monkeypatch.setenv("CNVRG_USER", "fake-username")
        monkeypatch.setenv("CNVRG_JWT_TOKEN", "fake-token")
        monkeypatch.setenv("CNVRG_ORGANIZATION", "fake-organization")

        assert empty_context._load_from_env()
        assert empty_context.token == "fake-token"
        assert empty_context.domain == "fake-domain"
        assert empty_context.user == "fake-username"
        assert empty_context.organization == "fake-organization"

    def test_load_from_pod_env(self, monkeypatch, empty_context):
        monkeypatch.setenv("CNVRG_URL", "fake-domain")
        monkeypatch.setenv("CNVRG_JWT_TOKEN", '')
        monkeypatch.setenv("CNVRG_ORGANIZATION", '')

        assert not empty_context._load_from_env()

        monkeypatch.setenv("CNVRG_URL", "fake-domain")
        monkeypatch.setenv("CNVRG_USER", "fake-username")
        monkeypatch.setenv("CNVRG_JWT_TOKEN", "fake-token")
        monkeypatch.setenv("CNVRG_ORGANIZATION", "fake-organization")

        # Pod env vars
        monkeypatch.setenv("CNVRG_PROJECT", "fake-project")
        monkeypatch.setenv("CNVRG_JOB_ID", "fake-job-slug")
        monkeypatch.setenv("CNVRG_JOB_TYPE", "Experiment")

        assert empty_context._load_from_env()
        assert empty_context.project == "fake-project"
        assert empty_context.experiment == "fake-job-slug"

    def test_load_from_config_file_doesnt_exist(self, mocker, empty_context):
        def mock_path_exists(*args, **kwargs):
            return False

        mocker.patch.object(os.path, "exists", mock_path_exists)

        assert not empty_context._load_from_config()



import pytest

from cnvrgv2.proxy import Proxy
from cnvrgv2.errors import CnvrgError
from cnvrgv2.modules.queries.query import Query


class TestQuery:
    def test_query_bad_init(self, empty_context):
        with pytest.raises(CnvrgError):
            Query(None)

    def test_delete_query(self, mocker, empty_context):
        slug = "test-slug"

        model = Query(context=empty_context, slug=slug)

        mocked_delete = mocker.patch.object(Proxy, 'call_api')
        model.delete()
        assert mocked_delete.call_count == 1



from datetime import datetime
from time import time

import botocore
from boto3 import Session
from boto3.s3.transfer import TransferConfig
from botocore.credentials import RefreshableCredentials
from botocore.session import get_session
from dateutil.tz import tzlocal

from cnvrgv2.data.clients.base_storage_client import BaseStorageClient
from cnvrgv2.utils.retry import retry
from cnvrgv2.utils.storage_utils import create_dir_if_not_exists

config = TransferConfig(max_concurrency=10, use_threads=True)


class S3Storage(BaseStorageClient):
    def __init__(self, refresh_function, storage_meta=None):
        """
        Inits a s3 client
        @param refresh_function: func. A function to retrieve storage credentials
        @param storage_meta: dict. Storage credentials. This is the output of the storage_function.
                We allow sending the first round of credentials to prevent double calls to the server at the initiation
        """
        if storage_meta is None:
            storage_meta = refresh_function()

        super().__init__(storage_meta)

        props = self._decrypt_dict(storage_meta, keys=["bucket", "region"])

        self.refresh_function = refresh_function
        self.bucket = props.get("bucket")
        self.region = props.get("region")
        self.client = self._get_client(storage_meta)

    @retry(log_error=True)
    def upload_single_file(self, local_path, object_path, progress_bar=None):
        try:
            self.client.upload_file(
                local_path,
                self.bucket,
                object_path,
                Config=config,
                Callback=self.progress_callback(progress_bar)
            )
        except Exception as e:
            print(e)
            raise e

    @retry(log_error=True)
    def download_single_file(self, local_path, object_path, progress_bar=None):
        try:
            create_dir_if_not_exists(local_path)
            if not object_path:
                return

            self.client.download_file(
                self.bucket,
                object_path,
                local_path,
                Config=config,
                Callback=self.progress_callback(progress_bar)
            )
        except Exception as e:
            raise e

    def _refresh(self, storage_meta=None):
        """
        Refreshes the credentials
        @param storage_meta: If sent, the refresh won't be called, the storage_meta will be assumed to be the server's
            response. This is to prevent double calling the server when starting the process
        @return:
        """
        props = self._decrypt_dict(
            storage_meta or self.refresh_function(),
            keys=["sts_a", "sts_s", "sts_st", "bucket", "region", "expiration"]
        )

        ttl = self._seconds_until_expiration(props.get("expiration"))

        credentials = {
            "access_key": props.get("sts_a"),
            "secret_key": props.get("sts_s"),
            "token": props.get("sts_st"),
            "expiry_time": datetime.fromtimestamp(time() + ttl, tz=tzlocal()).isoformat()
        }
        return credentials

    @staticmethod
    def _seconds_until_expiration(datetime_string, datetime_format="%Y-%m-%d %H:%M:%S %Z"):
        """
        Calculates the number of seconds until the given expiration time
        @param datetime_string: string. A datetime string
        @param datetime_format: string. Format of the given datetime_string. defaults to the format returned from
            cnvrg server
        @return: Seconds until the given datetime
        """
        if not datetime_string:
            # 36 hours, this is the default for ci-cd in the server.
            # Add here as a fallback for old server versions that doesn't return an expiration
            return 129600
        try:
            # In case the given string is an integer
            return int(datetime_string)
        except ValueError:
            # In case it's a string representing a datetime
            target_time = datetime.strptime(datetime_string, datetime_format)
            now = datetime.utcnow()
            time_delta = target_time - now
            return time_delta.total_seconds()

    def _get_client(self, storage_meta):
        """
        Generates a client fo the process using a RefreshableCredentials session.
        @param storage_meta: dict. Credentials to begin the process with
        @return: A boto3 client
        """
        session_credentials = RefreshableCredentials.create_from_metadata(
            metadata=self._refresh(storage_meta=storage_meta),
            refresh_using=self._refresh,
            method="sts_assume_role"
        )
        session = get_session()
        session._credentials = session_credentials
        session.set_config_variable("region", self.region)
        autorefresh_session = Session(botocore_session=session)

        botocore_config = botocore.config.Config(max_pool_connections=50)
        return autorefresh_session.client('s3', config=botocore_config, verify=self.check_certificate)



from cnvrgv2.proxy import Proxy
from cnvrgv2.modules.project import Project
from cnvrgv2.modules.project_settings import ProjectSettings


class TestProjectSettings:

    def test_save_project_settings(self, mocker, empty_context):
        project = Project(context=empty_context, slug="fake-slug")
        project_settings = ProjectSettings(project=project)

        mocked_save = mocker.patch.object(Proxy, 'call_api')
        project_settings.save()
        assert mocked_save.call_count == 1



import pytest

from cnvrgv2 import Cnvrg
from cnvrgv2.errors import CnvrgError, CnvrgHttpError, CnvrgLoginError
from cnvrgv2.modules.organization.organizations_client import OrganizationsClient
from cnvrgv2.modules.users import ROLES
from cnvrgv2.modules.users.users_client import UsersClient
from tests.e2e.conftest import call_database


class TestUsers:
    @staticmethod
    def cleanup(context_prefix):
        delete_user_command = "DELETE FROM users WHERE username LIKE'{}%'".format(context_prefix)
        call_database(delete_user_command)

    def test_create_user(self, class_context, domain, e2e_env):
        username = class_context.generate_name(5)
        email = username + "@cnvrg.io"
        password = "qwe123"

        uc = UsersClient(domain=domain)
        user = uc.register(username=username, email=email, password=password)
        assert user

    def test_user_update_success(self, domain, e2e_user):
        uc = UsersClient(domain=domain)
        uc.login(user=e2e_user["email"], password=e2e_user["password"])

        user = uc.me()
        new_name = "new_name"
        new_git_access_token = "larger_than_eight_characters"
        new_company = "new_company"
        new_bio = "new_bio"
        new_timezone = "Asia/Jerusalem"

        # Test single update
        user.name = new_name
        user.save()

        # Test multi update
        user.update(
            **{
                "git_access_token": new_git_access_token,
                "company": new_company,
                "bio": new_bio,
                "time_zone": new_timezone
            }
        )

        assert user.name == new_name
        assert user.git_access_token == new_git_access_token[-8:]
        assert user.company == new_company
        assert user.bio == new_bio
        assert user.time_zone == new_timezone

    def test_create_duplicate_user(self, class_context, domain, e2e_env):
        username = class_context.generate_name(5)
        email = username + "@cnvrg.io"
        password = "qwe123"

        uc = UsersClient(domain=domain)
        user = uc.register(username=username, email=email, password=password)
        assert user

        with pytest.raises(CnvrgHttpError) as exception_info:
            uc.register(username=username, email=email, password=password)
        assert "exists" in str(exception_info.value)

    def test_create_user_bad_arguments(self, domain, e2e_env):
        username = "bad"
        email = username
        password = "qwe123"

        uc = UsersClient(domain=domain)
        with pytest.raises(CnvrgError) as exception_info:
            uc.register(username=username, email=email, password=password)
        assert "Faulty" in str(exception_info.value)

    def test_user_login(self, class_context, domain, e2e_env):
        rand_name = class_context.generate_name(5)
        user_info = {
            "username": rand_name,
            "email": rand_name + "@cnvrg.io",
            "password": "qwe123"
        }

        uc = UsersClient(domain=domain)
        uc.register(**user_info)
        token, org, _ = uc.login(user=user_info["email"], password=user_info["password"])
        assert type(token) == str
        assert org is None

    def test_user_login_bad_user(self, domain, e2e_env):
        uc = UsersClient(domain=domain)
        with pytest.raises(CnvrgLoginError) as exception_info:
            uc.login(user="fake@fake.com", password="fake")
        assert "login" in str(exception_info.value)

    def test_user_me(self, class_context, domain, e2e_env):
        username = class_context.generate_name(5)
        email = username + "@cnvrg.io"
        password = "qwe123"

        uc = UsersClient(domain=domain)
        uc.register(username=username, email=email, password=password)
        uc.login(user=email, password=password)

        user = uc.me()
        assert user.email == email
        assert user.username == username

    def test_user_leave_org(self, class_context, domain, e2e_env):
        username = class_context.generate_name(5)
        email1 = username + "1@cnvrg.io"
        email2 = username + "2@cnvrg.io"
        password = "qwe123"

        uc = UsersClient(domain=domain)
        user1 = uc.register(username=username + "1", email=email1, password=password)
        user2 = uc.register(username=username + "2", email=email2, password=password)

        assert len(user1.organizations) == 0
        assert len(user2.organizations) == 0

        org_name = username + "org"
        oc = OrganizationsClient(domain=domain, token=user1._token)
        oc.create(name=org_name)

        user1.reload()
        user2.reload()
        assert len(user1.organizations) == 1
        assert len(user2.organizations) == 0

        with pytest.raises(CnvrgHttpError) as exception_info:
            user1.leave_org(org_name)
        assert "you created" in str(exception_info.value)

        cnvrg = Cnvrg(domain=domain, email=email1, password=password)
        cnvrg.members.add(email=email2, role=ROLES.MANAGER)
        user2.reload()
        assert len(user1.organizations) == 1

        # Make sure we remove the organization membership both locally and in server (tested after reload)
        user2.leave_org(org_name)
        assert len(user2.organizations) == 0
        user2.reload()
        assert len(user2.organizations) == 0



import click

import os.path
from cnvrgv2.cnvrg import UsersClient
from cnvrgv2.cli.utils import messages
from cnvrgv2.config import Config

from cnvrgv2.errors import CnvrgFileError, CnvrgHttpError
from cnvrgv2.cli.utils.decorators import prepare_command_without_context
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.config import error_messages


@click.command()
@click.option('-d', '--domain', default=None, help=messages.LOGIN_PROMPT_DOMAIN)
@click.option('-e', '--email', default=None, help=messages.LOGIN_PROMPT_EMAIL)
@click.option('-p', '--password', default=None, help=messages.LOGIN_PROMPT_PASSWORD, hide_input=True)
@click.option('-o', '--organization', default=None, help=messages.LOGIN_ORGANIZATION_HELP)
@click.option('-t', '--auth-token', default=None, help=messages.LOGIN_HELP_AUTH_TOKEN)
@prepare_command_without_context()
def login(logger, domain, email, password, organization, auth_token):
    config = Config()
    config_domain = None
    if password is None:
        password = ""

    if domain is None:
        config_domain = config.domain
        domain = config_domain or click.prompt(messages.LOGIN_PROMPT_DOMAIN, type=str)

    # The guideline in this block is to receive both email and pass together from the user or config file.
    if email is None:
        config_email = config.user
        config_token = config.token

        if config_email and config_token and config_domain:
            click.echo(messages.LOGIN_ALREADY_LOGGED_IN)
            return
        else:
            email = click.prompt(messages.LOGIN_PROMPT_EMAIL, type=str)
            if not auth_token:
                password = click.prompt(messages.LOGIN_PROMPT_PASSWORD, type=str, hide_input=True)
    else:
        if not auth_token:
            password = password or click.prompt(messages.LOGIN_PROMPT_PASSWORD, type=str, hide_input=True)

    try:
        uc = UsersClient(domain=domain, token=auth_token)
        config.sso_version = uc.sso_version

        token, default_organization, sso_enabled = uc.login(user=email, password=password, token=auth_token)

        config.domain = domain
        config.organization = organization or default_organization
        config.user = email
        if auth_token is not None and sso_enabled:
            config.token = auth_token
        else:
            config.token = token
        config.check_certificate = False
        config.save()

        logger.log_and_echo(messages.LOGIN_SUCCESS.format(email))

    except CnvrgHttpError:
        logger.log_and_echo(messages.LOGIN_INVALID_CREDENTIALS, error=True)


@click.command()
@prepare_command_without_context()
def logout(logger):
    try:
        config = Config()
        config.remove_config_fields("user", "token")
        logger.log_and_echo(messages.LOGOUT_SUCCESS)

    except CnvrgFileError:
        logger.log_and_echo(messages.LOGOUT_CONFIG_MISSING, error=True)


@click.command()
@prepare_command()
def me(logger):
    """
    Prints the current logged in user details
    """
    try:
        config = Config()
        log_dir = os.path.join(config.global_cnvrg_path, "logs")
        log_path = log_dir + "/cnvrg.log"
        click.echo(messages.ME_SUCCESS_API.format(config.domain))
        click.echo(messages.ME_SUCCESS.format(config.user))
        click.echo(messages.ME_SUCCESS_LOGS.format(log_path))
        logger.info(messages.ME_LOGGER_SUCCESS)

    except CnvrgFileError:
        logger.log_and_echo(error_messages.CREDENTIALS_MISSING, error=True)



import pytz

from cnvrgv2.config import error_messages
from cnvrgv2.errors import CnvrgArgumentsError


class Cron:
    def __init__(self, minute="*", hour="*", day_of_month="*", month="*", day_of_week="*", timezone=None):
        Cron.validate_input(locals())
        self._minute = minute
        self._hour = hour
        self._day_of_month = day_of_month
        self._month = month
        self._day_of_week = day_of_week
        self._timezone = timezone

    @staticmethod
    def validate_input(cron_params):
        # type check
        allowed_values = {
            "minute": {"min": 0, "max": 59},
            "hour": {"min": 0, "max": 23},
            "day_of_month": {"min": 1, "max": 31},
            "month": {"min": 1, "max": 12},
            "day_of_week": {"min": 0, "max": 6}
        }

        argument_errors = {}
        for param_name, edges in allowed_values.items():
            current_value = cron_params[param_name]
            if current_value == "*":
                continue

            if not isinstance(current_value, int) or current_value not in range(edges["min"], edges["max"] + 1):
                message = error_messages.INVALID_CRON_ARGUMENT.format(param_name, edges["min"], edges["max"])
                argument_errors[param_name] = message

        timezone = cron_params.get("timezone", None)
        if timezone and timezone not in pytz.all_timezones:
            argument_errors["timezone"] = error_messages.CRON_INVALID_TIMEZONE

        if argument_errors:
            raise CnvrgArgumentsError(argument_errors)

    @staticmethod
    def validate_cron_syntax(cron_string):
        cron_parts = cron_string.strip().split()
        if len(cron_parts) < 5:
            raise CnvrgArgumentsError(error_messages.CRON_ARGUMENTS_MISSING)

        cron_params = {
            "minute": cron_parts[0],
            "hour": cron_parts[1],
            "day_of_month": cron_parts[2],
            "month": cron_parts[3],
            "day_of_week": cron_parts[4],
            "timezone": cron_parts[5] if 5 < len(cron_parts) else None
        }

        try:
            for key, value in cron_params.items():
                if key == "timezone":
                    continue
                if value != "*":
                    cron_params[key] = int(value)
        except ValueError:
            raise CnvrgArgumentsError(error_messages.CRON_INVALID_ARGUMENT)

        Cron.validate_input(cron_params)

    def __str__(self):
        cron_string = "{} {} {} {} {}".format(
            self._minute,
            self._hour,
            self._day_of_month,
            self._month,
            self._day_of_week
        )

        if self._timezone is not None:
            cron_string += " " + self._timezone

        return cron_string



from cnvrgv2.context import Context
from cnvrgv2.modules.ssh.ssh import Ssh


class SshClient:
    def __init__(self, organization):
        self._context = Context(context=organization._context)

    def create_ssh(self, job_id):
        return Ssh(self._context, job_id)



import pytest

from cnvrgv2.config.config import SSO_VERSION
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.users.user import User
from cnvrgv2.modules.users.users_client import UsersClient
from cnvrgv2.proxy import Proxy
from cnvrgv2.utils.json_api_format import JAF


class TestUsersClient:
    def test_user_client_invalid_domain(self):
        domain = "bad_domain"
        with pytest.raises(CnvrgArgumentsError) as exception_info:
            UsersClient(domain=domain)

        assert "URL" in str(exception_info.value)

    def test_user_client_valid_domain(self, mocker):
        def mock_get_sso_version(*args, **kwargs):
            return SSO_VERSION.CAPI

        mocker.patch.object(UsersClient, "_get_sso_version", side_effect=mock_get_sso_version)

        domain = "http://localhost:3000"
        UsersClient(domain=domain)

        domain = "http://0.0.0.0:3000"
        UsersClient(domain=domain)

        domain = "https://example.com"
        UsersClient(domain=domain)

        assert True

    def test_user_register_bad_email(self, mocker):
        def mock_call_api(*args, **kwargs):
            return

        def mock_get_sso_version(*args, **kwargs):
            return SSO_VERSION.CAPI

        mock_response = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        mocker.patch.object(UsersClient, "_get_sso_version", side_effect=mock_get_sso_version)

        domain = "http://localhost:3000"
        uc = UsersClient(domain=domain)

        with pytest.raises(CnvrgArgumentsError) as exception_info:
            uc.register(username="valid_user", email="invalid email", password="password")

        assert "email" in str(exception_info.value)
        assert mock_response.call_count == 0

    def test_user_register_invalid_character_in_domain(self, mocker):
        def mock_call_api(*args, **kwargs):
            return

        def mock_get_sso_version(*args, **kwargs):
            return SSO_VERSION.CAPI

        # Mock the call_api and _get_sso_version methods
        mock_response = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        mocker.patch.object(UsersClient, "_get_sso_version", side_effect=mock_get_sso_version)

        domain = "http://localhost:3000"
        uc = UsersClient(domain=domain)

        # Test with invalid email containing forbidden character
        with pytest.raises(CnvrgArgumentsError):
            uc.register(username="invalid_user", email="nonexistent@-example.com", password="password")

        assert mock_response.call_count == 0

    def test_user_register_bad_username(self, mocker):
        def mock_call_api(*args, **kwargs):
            return

        def mock_get_sso_version(*args, **kwargs):
            return SSO_VERSION.CAPI

        mock_response = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        mocker.patch.object(UsersClient, "_get_sso_version", side_effect=mock_get_sso_version)

        domain = "http://localhost:3000"
        uc = UsersClient(domain=domain)

        with pytest.raises(CnvrgArgumentsError) as exception_info:
            uc.register(username="invalid user", email="valid@email.com", password="password")

        assert "user" in str(exception_info.value)
        assert mock_response.call_count == 0

    def test_user_register_valid_inputs(self, mocker):
        def mock_call_api(*args, **kwargs):
            return JAF(response={
                "data": {
                    "id": "1",
                    "type": "user",
                    "attributes": {
                        "email": "valid@email.com",
                        "username": "valid_user",
                    }
                },
                "meta": {
                    "jwt": "asdasd"
                }

            })

        def mock_get_sso_version(*args, **kwargs):
            return SSO_VERSION.CAPI

        mock_response = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        mocker.patch.object(UsersClient, "_get_sso_version", side_effect=mock_get_sso_version)

        domain = "http://localhost:3000"
        uc = UsersClient(domain=domain)

        user = uc.register(username="validuser", email="val.ID@email.com", password="password")
        assert isinstance(user, User)
        assert mock_response.call_count == 1



import click

'''
Parser for dictionary-like cli parameter
usage example:
python cli.py --my-dict-option key1=value1,key2=value2
'''


class DictParamType(click.ParamType):
    name = 'dictionary'

    def convert(self, value, param, ctx):
        try:
            items = value.split(',')
            return dict(item.split('=') for item in items)
        except Exception:
            self.fail(f'{value} is not a valid dictionary format', param, ctx)


DICT = DictParamType()



import re

TAG_VALIDATION_REGEX = re.compile(r"\A[\w][\w.\-]{0,127}\Z")
NAME_VALIDATION_REGEX = re.compile(r"\A[a-z0-9\-_\/]+\Z")


class ImageLogos:
    KERAS = "keras"
    TENSORFLOW = "tensorflow"
    SKLEARN = "sklearn"
    PYTHON = "python"
    R = "r"
    XGBOOST = "xgboost"
    BASH = "bash"
    S3 = "s3"
    SPARK = "spark"
    OPENCV = "opencv"
    PYTORCH = "pytorch"
    VGG = "vgg"
    CNVRG = "cnvrg"
    MXNET = "mxnet"
    TENSOR_RT = "tensor_rt"
    RAPIDS = "rapids"
    INTEL = "intel"
    NVIDIA = "nvidia"

    @classmethod
    def validate_icon(cls, icon_name):
        if icon_name is None or icon_name == "":
            return True

        icon_types = [cls.__dict__[var] for var in vars(cls) if not var.startswith("__")]
        return icon_name in icon_types



import os
import shutil
import pytest

from os import path
from tests.e2e.conftest import call_database

from cnvrgv2.errors import CnvrgHttpError
from cnvrgv2.modules.dataset import Dataset


class TestDatasets:

    @staticmethod
    def cleanup(context_prefix):
        delete_user_command = "DELETE FROM datasets WHERE slug LIKE'{}%'".format(context_prefix)
        call_database(delete_user_command)

    def test_create(self, class_context, e2e_client):
        name = class_context.generate_name(5)
        model = e2e_client.datasets.create(name=name)
        assert type(model) == Dataset
        assert model.slug == name

    def test_get(self, class_context, e2e_client):
        name = class_context.generate_name(5)
        e2e_client.datasets.create(name=name)

        model = e2e_client.datasets.get(slug=name)
        assert type(model) == Dataset
        assert model.slug == name
        assert model.title == name

    def test_get_non_existent(self, class_context, e2e_client):
        name = class_context.generate_name(5)

        model = e2e_client.datasets.get(slug=name)
        assert type(model) == Dataset

        with pytest.raises(CnvrgHttpError) as exception_info:
            model.title

        assert "found" in str(exception_info.value)

    def test_delete(self, class_context, e2e_client):
        name = class_context.generate_name(5)
        e2e_client.datasets.create(name=name)

        model = e2e_client.datasets.get(slug=name)
        assert model.title == name

        model.delete()

        model = e2e_client.datasets.get(slug=name)
        with pytest.raises(CnvrgHttpError) as exception_info:
            model.title

        assert "found" in str(exception_info.value)

    def test_delete_non_existent(self, class_context, e2e_client):
        name = class_context.generate_name(5)

        model = e2e_client.datasets.get(slug=name)
        with pytest.raises(CnvrgHttpError) as exception_info:
            model.delete()

        assert "found" in str(exception_info.value)

    def test_list(self, class_context, e2e_client):
        # Clean the db before testing list
        TestDatasets.cleanup(class_context.prefix)

        name_base = class_context.generate_name(5)
        for i in range(40):
            name = "{}{}".format(name_base, i)
            e2e_client.datasets.create(name=name)

        # Check descending order works
        idx = 39
        models = e2e_client.datasets.list()
        for model in models:
            assert model.slug == "{}{}".format(name_base, idx)
            idx -= 1

        # Check ascending order works
        idx = 0
        models = e2e_client.datasets.list(sort="id")
        for model in models:
            assert model.slug == "{}{}".format(name_base, idx)
            idx += 1

    def test_search(self, class_context, e2e_client):
        # Clean the db before testing list
        TestDatasets.cleanup(class_context.prefix)

        name_base = class_context.generate_name(5)
        search_slugs = []

        for i in range(5):
            name = "{}{}".format(name_base, i)
            e2e_client.datasets.create(name=name)

        models = e2e_client.datasets.list()
        for model in models:
            search_slugs.append(model.slug)

        for slug in search_slugs:
            datasets = e2e_client.datasets.search(slug)
            for dataset in datasets:
                assert dataset.slug == slug

    def test_clone_success(self, random_name, e2e_dataset):
        try:
            # Arrange
            filename = random_name(5)
            expected_output_dir = path.join(e2e_dataset.title)
            expected_output_file = path.join(expected_output_dir, filename)
            open(filename, 'w')
            e2e_dataset.put_files([filename], message="test dataset files")

            e2e_dataset.clone()
            assert path.exists(expected_output_file)
        finally:
            # cleanup
            if path.exists(filename):
                os.remove(filename)
            if path.exists(expected_output_dir):
                shutil.rmtree(expected_output_dir)



from unittest.mock import patch
from tests.unit.modules.datasource.conftest import file_path


class TestDatasourceUpload:
    destination_path = 'destination_path'

    @patch('os.path.basename')
    def test_upload(self, mock_basename, empty_context, datasource, mock_datasource_client, mock_normalize_file_path):
        datasource.upload_file(file_path=file_path, destination_path=self.destination_path)

        datasource._normalize_file_path.assert_called_once_with(self.destination_path)
        datasource.client.upload_file.assert_called_once_with(
            Bucket=datasource.bucket_name,
            Filename=file_path,
            Key=mock_normalize_file_path.return_value
        )

    @patch('os.path.basename')
    def test_upload_destination_none(self, mock_basename, empty_context, datasource, mock_datasource_client,
                                     mock_normalize_file_path):
        datasource.upload_file(file_path=file_path, destination_path=None)

        datasource._normalize_file_path.assert_called_once_with(file_path)
        datasource.client.upload_file.assert_called_once_with(
            Bucket=datasource.bucket_name,
            Filename=file_path,
            Key=mock_normalize_file_path.return_value
        )

    @patch('os.path.basename')
    def test_upload_destination_not_provided(self, mock_basename, empty_context, datasource, mock_datasource_client,
                                             mock_normalize_file_path):
        datasource.upload_file(file_path=file_path)

        datasource._normalize_file_path.assert_called_once_with(file_path)
        datasource.client.upload_file.assert_called_once_with(
            Bucket=datasource.bucket_name,
            Filename=file_path,
            Key=mock_normalize_file_path.return_value
        )

    @patch('os.path.basename')
    def test_upload_destination_empty_str(self, mock_basename, empty_context, datasource, mock_datasource_client,
                                          mock_normalize_file_path):
        datasource.upload_file(file_path=file_path, destination_path='')

        datasource._normalize_file_path.assert_called_once_with(file_path)
        datasource.client.upload_file.assert_called_once_with(
            Bucket=datasource.bucket_name,
            Filename=file_path,
            Key=mock_normalize_file_path.return_value
        )



from cnvrgv2.proxy import Proxy
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.config import error_messages, routes
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.modules.resources._spark_driver import SparkDriver


class SparkDriversClient:
    def __init__(self, organization):
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.SPARK_DRIVERS_BASE.format(scope["organization"])

    def list(self, sort="-id"):
        """
        List all spark drivers in a specific resource
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @raise: HttpError
        @return: Generator that yields project objects
        """
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=SparkDriver,
            sort=sort
        )

    def get(self, slug):
        """
        Retrieves a spark driver by the given slug
        @param slug: The slug of the requested spark driver
        @return: Spark driver object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.SPARK_DRIVER_GET_FAULTY_SLUG)

        return SparkDriver(context=self._context, slug=slug)

    # TODO: Add create function



import pytest

from cnvrgv2.cnvrg import Cnvrg
from cnvrgv2.errors import CnvrgError
from cnvrgv2.modules.organization.organizations_client import OrganizationsClient
from cnvrgv2.modules.users import ROLES
from cnvrgv2.modules.users.users_client import UsersClient
from tests.e2e.conftest import call_database


class TestCnvrg:
    @staticmethod
    def cleanup(context_prefix):
        delete_user_command = "DELETE FROM organizations WHERE slug LIKE'{}%'".format(context_prefix)
        call_database(delete_user_command)

    def test_me(self, domain, e2e_client):
        user_mail = e2e_client._context.user
        ret_user = e2e_client.me()
        assert ret_user.email == user_mail

    def test_switch_organizations(self, class_context, domain, e2e_client):
        alter_org = class_context.generate_name(5)
        original_org_proj = class_context.prefix + "orig-proj"
        alter_org_proj = class_context.prefix + "alter-proj"

        oc = OrganizationsClient(domain=domain, token=e2e_client._context.token)
        oc.create(alter_org)

        e2e_client.projects.create(name=original_org_proj)
        assert e2e_client.projects.get(original_org_proj).slug == original_org_proj

        e2e_client.set_organization(alter_org)
        e2e_client.projects.create(alter_org_proj)

        assert e2e_client._context.organization == alter_org
        assert e2e_client.projects.get(alter_org_proj).slug == alter_org_proj

    def test_get_users(self, domain, e2e_client, class_context):
        # e2e_client has at least 2 users here: 1 added when its created,
        # 2nd is created in this test
        username = class_context.generate_name(5)
        email = username + "@cnvrg.io"
        password = "qwe123"
        uc = UsersClient(domain=domain)
        uc.register(username=username, email=email, password=password)
        e2e_client.members.add(email=email, role=ROLES.MEMBER)
        users = e2e_client.get_users()
        assert len(users) > 1

    def test_get_users_unauthorized(self, domain, e2e_client, e2e_data_scientist_user):
        client_of_non_admin_user = Cnvrg(
            domain=domain,
            email=e2e_data_scientist_user["email"],
            password=e2e_data_scientist_user["password"],
            organization=e2e_client._context.organization
        )
        with pytest.raises(CnvrgError) as exception_info:
            client_of_non_admin_user.get_users()

        assert 'unauthorized' in str(exception_info.value).lower()



import pytest

from cnvrgv2.proxy import Proxy
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.modules.organization.organizations_client import OrganizationsClient


class TestOrganizationsClient:
    def test_init_valid_credentials(self):
        organizations_client = OrganizationsClient(domain="fake", token="fake")
        assert organizations_client._proxy

    def test_get_organization_with_slug(self, mocker):
        slug = "test-slug"

        def mock_call_api(*args, **kwargs):
            response = {
                "data": {
                    "id": "2",
                    "type": "organization",
                    "attributes": {
                        "slug": slug,
                        "created_at": "2020-09-13T15:32:16.671Z",
                        "updated_at": "2020-09-13T15:32:16.671Z"
                    }
                }
            }
            return JAF(response=response)

        mocked_api_call = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        organizations_client = OrganizationsClient(domain="fake", token="fake")
        organizations_client.get(slug)

        # TODO: add attribute checks. added call_count until decided what to check
        assert mocked_api_call.call_count > 0

    def test_get_organization_empty_slug(self, empty_context):
        organizations_client = OrganizationsClient(empty_context.domain, empty_context.token)
        with pytest.raises(CnvrgArgumentsError):
            organizations_client.get("")

    def test_get_organization_none_slug(self, empty_context):
        organizations_client = OrganizationsClient(empty_context.domain, empty_context.token)
        with pytest.raises(CnvrgArgumentsError):
            organizations_client.get(None)

    def test_get_organization_obj_slug(self, empty_context):
        organizations_client = OrganizationsClient(empty_context.domain, empty_context.token)
        with pytest.raises(CnvrgArgumentsError):
            organizations_client.get({"say": "whaaat??"})

    def test_create_organization_with_name(self, mocker, empty_context):
        slug = "test-slug"

        def mock_call_api(*args, **kwargs):
            response = {
                "data": {
                    "id": "2",
                    "type": "organization",
                    "attributes": {
                        "slug": slug,
                        "created_at": "2020-09-13T15:32:16.671Z",
                        "updated_at": "2020-09-13T15:32:16.671Z"
                    }
                }
            }
            return JAF(response)

        mocked_api_call = mocker.patch.object(Proxy, "call_api", side_effect=mock_call_api)
        organizations_client = OrganizationsClient(empty_context.domain, empty_context.token)
        organizations_client.create(slug)

        # TODO: add attribute checks. added call_count until decided what to check
        assert mocked_api_call.call_count > 0

    def test_create_organization_empty_name(self, empty_context):
        organizations_client = OrganizationsClient(empty_context.domain, empty_context.token)
        with pytest.raises(CnvrgArgumentsError):
            organizations_client.create("")

    def test_create_organization_none_name(self, empty_context):
        organizations_client = OrganizationsClient(empty_context.domain, empty_context.token)
        with pytest.raises(CnvrgArgumentsError):
            organizations_client.create(None)

    def test_create_organization_obj_name(self, empty_context):
        organizations_client = OrganizationsClient(empty_context.domain, empty_context.token)
        with pytest.raises(CnvrgArgumentsError):
            organizations_client.create({"say": "whaaat??"})



import pytest


class TestDatasourceNormalizePath:

    # Run the tests multiple times, with different params
    @pytest.mark.parametrize(
        "datasource_path, file_name, expected_file_name",
        [
            # Test case 1 file_name
            # case 1
            ('path1', 'file_name', 'path1/file_name'),
            # case 2
            ('path2/', 'file_name', 'path2/file_name'),
            # case 3
            ('', 'file_name', 'file_name'),
            # Test case 2 path/to/file_name (relative path)
            # case 4
            ('path1', 'path/to/file_name', 'path1/path/to/file_name'),
            # case 5
            ('path2/', 'path/to/file_name', 'path2/path/to/file_name'),
            # case 6
            ('', 'path/to/file_name', 'path/to/file_name'),
            # Test case 3 /path/to/file_name (absolute path)
            # case 7
            ('path1', '/path/to/file_name', 'path1/path/to/file_name'),
            # case 8
            ('path2/', '/path/to/file_name', 'path2/path/to/file_name'),
            # case 9
            ('', '/path/to/file_name', '/path/to/file_name'),
            # Test case 4 datasource path: '/'
            # case 10
            ('/', 'file_name', '/file_name'),
            # case 11
            ('/', 'path/to/file_name', '/path/to/file_name'),
            # case 12
            ('/', '/path/to/file_name', '/path/to/file_name'),
            # Test case 5 path1/path/to/file_name (path already containing path)
            # case 13
            ('path1', 'path1/path/to/file_name', 'path1/path1/path/to/file_name'),
            # case 14
            ('path2/', 'path2/path/to/file_name', 'path2/path2/path/to/file_name'),

        ],
        ids=[
            # Test case 1 file_name
            "1 - path1 & file_name",
            "2 - path2 & file_name",
            "3 - no path & file_name",
            # Test case 2 path/to/file_name
            "4 - path1 & path/to/file_name",
            "5 - path2 & path/to/file_name",
            "6 - no path & path/to/file_name",
            # Test case 3 /path/to/file_name (absolute path)
            "7 - path1 & /path/to/file_name",
            "8 - path2 & /path/to/file_name",
            "9 - no path & /path/to/file_name",
            # Test case 4 /path/to/file_name (absolute path)
            "10 - / & file_name",
            "11 - / & path/to/file_name",
            "12 - / & /path/to/file_name",
            "13 - / & path1/path/to/file_name",
            "14 - / & path2/path/to/file_name",
             ]
    )
    def test_normalize_file_path(self, datasource, mock_datasource_client,
                                 datasource_path,  file_name, expected_file_name):
        datasource.path = datasource_path
        normalized_file_name = datasource._normalize_file_path(file_path=file_name)
        assert normalized_file_name == expected_file_name



import os

import pytest

from tests.e2e.data_owner_e2e.data_owner_files import DataOwnerFiles
from tests.e2e.test_queries import TestQueries


class TestDatasetFiles(DataOwnerFiles):

    def test_put_files_success(self, e2e_dataset, temp_file):
        self.data_owner_put_files(e2e_dataset, temp_file)

    def test_put_unchanged_files_success(self, e2e_dataset, temp_file):
        self.data_owner_put_unchanged_files(e2e_dataset, temp_file)

    def test_put_folder_success(self, e2e_dataset, tmpdir):
        self.data_owner_put_folder(e2e_dataset, tmpdir)

    def test_put_files_override_false_success(self, e2e_dataset, temp_file):
        self.data_owner_put_files_override_false(e2e_dataset, temp_file)

    def test_put_files_override_true_success(self, e2e_dataset, temp_file):
        self.data_owner_put_files_override_true(e2e_dataset, temp_file)

    def test_put_file_abs_path_dir_path_success(self, e2e_dataset, temp_file):
        self.data_owner_put_files_abs_path_dir_path(e2e_dataset, temp_file)

    def test_put_file_dir_path_relative_path_success(self, e2e_dataset, temp_file):
        self.data_owner_put_file_dir_path_relative_path(e2e_dataset, temp_file)

    def test_put_folder_abs_path_dir_path_success(self, e2e_dataset, tmpdir):
        self.data_owner_put_folder_abs_path_dir_path(e2e_dataset, tmpdir)

    def test_put_folder_dir_path_relative_path_success(self, e2e_dataset, tmpdir):
        self.data_owner_put_folder_dir_path_relative_path(e2e_dataset, tmpdir)

    def test_put_empty_folder_dir_path_relative_path_success(self, e2e_dataset, tmpdir):
        self.data_owner_put_folder_dir_path_relative_path(e2e_dataset, tmpdir)

    def test_put_files_force_success(self, e2e_dataset, temp_file, temp_file_second):
        self.data_owner_put_files_force(e2e_dataset, temp_file, temp_file_second)

    def test_list_files_success(self, e2e_dataset, tmpdir):
        self.data_owner_list_files(e2e_dataset, tmpdir)

    def test_list_files_with_names_success(self, e2e_dataset, tmpdir):
        for num_files in [5, 20, 30, 60]:
            self.data_owner_list_files_var(e2e_dataset, tmpdir, num_files)

    def test_list_files_and_folders_success(self, e2e_dataset, tmpdir):
        self.data_owner_list_files_and_folders(e2e_dataset, tmpdir)

    def test_delete_files_success(self, e2e_dataset, temp_file):
        self.data_owner_delete_files(e2e_dataset, temp_file)

    def test_clone_files_success(self, e2e_dataset, tmpdir):
        self.data_owner_clone_files(e2e_dataset, tmpdir)

    def test_clone_with_empty_folder_success(self, e2e_dataset, tmpdir):
        self.data_owner_clone_with_empty_folder(e2e_dataset, tmpdir)

    def test_clone_files_override_success(self, e2e_dataset, tmpdir, temp_file):
        self.data_owner_clone_override(e2e_dataset, tmpdir, temp_file)

    def test_upload_files_success(self, e2e_dataset, tmpdir):
        self.data_owner_upload_files(e2e_dataset, tmpdir)

    def test_upload_unchanged_files_success(self, e2e_dataset, tmpdir):
        self.data_owner_upload_unchanged_files(e2e_dataset, tmpdir)

    def test_upload_files_from_non_cnvrg_fails(self, e2e_dataset, tmpdir):
        self.data_owner_upload_files_from_non_cnvrg(e2e_dataset, tmpdir)

    def test_download_files_success(self, e2e_dataset):
        self.data_owner_download_files(e2e_dataset)

    def test_download_empty_folder_success(self, tmpdir, e2e_dataset):
        self.data_owner_download_empty_folder(e2e_dataset, tmpdir)

    def test_put_files_absolute_path_success(self, tmpdir, e2e_dataset):
        self.data_owner_put_files_absolute_path(tmpdir, e2e_dataset)

    def test_remove_files_absolute_path_success(self, tmpdir, e2e_dataset):
        self.data_owner_remove_files_absolute_path(tmpdir, e2e_dataset)

    def test_remove_folders_through_upload_success(self, e2e_dataset, tmpdir):
        self.data_owner_remove_folders_through_upload(tmpdir, e2e_dataset)

    def test_remove_file_through_upload(self, e2e_dataset, tmpdir):
        self.data_owner_remove_file_through_upload(tmpdir, e2e_dataset)

    def test_list_commits_success(self, e2e_dataset, temp_file):
        self.data_owner_list_commits(e2e_dataset, temp_file)

    def test_access_all_commit_fields_success(self, e2e_dataset, temp_file, second_temp_file):
        self.data_owner_access_all_commit_fields(e2e_dataset, temp_file, second_temp_file)

    def test_remove_folder_locally_through_download_success(self, e2e_dataset, tmpdir):
        self.data_owner_remove_folder_locally_through_download(tmpdir, e2e_dataset)

    def test_remove_file_locally_through_download_success(self, e2e_dataset, tmpdir):
        self.data_owner_remove_file_locally_through_download(tmpdir, e2e_dataset)

    def test_data_owner_nested(self, e2e_dataset, tmpdir):
        self.data_owner_nested(tmpdir, e2e_dataset)

    def test_data_owner_cache(self, e2e_dataset, tmpdir):
        self.data_owner_cache(tmpdir, e2e_dataset)

    def test_data_owner_verify_cnvrg_files_exists_success(self, e2e_dataset, tmpdir):
        self.data_owner_verify_cnvrg_files_exists(e2e_dataset, tmpdir)

    def test_sync_not_implemented(self, e2e_dataset):
        with pytest.raises(NotImplementedError):
            e2e_dataset.sync()

    def test_clone_txt_files_using_query_success(self, e2e_dataset, tmpdir):
        # Inflate dataset with files. Half of the files will have a .txt suffix and will be downloaded using query
        num_of_files = 100  # Must be even
        os.chdir(tmpdir.strpath)
        test_dir_name = "test_directory"
        os.mkdir(test_dir_name)
        for i in range(num_of_files):
            filename = "file_{}".format(i)
            if i % 2 == 0:
                filename = filename + ".txt"
            with open(os.path.join(test_dir_name, filename), "w") as f:
                f.write(str(i))

        files_to_upload = "{}/*".format(test_dir_name)
        e2e_dataset.put_files(paths=[files_to_upload])

        # Create query and clone
        query = TestQueries.create_query(e2e_dataset, "txt-query", "*txt")
        e2e_dataset.clone(query_slug=query.slug)

        # Assertions
        assert len(os.listdir(os.path.join(e2e_dataset.slug, test_dir_name))) == num_of_files / 2

    def test_clone_empty_folder_using_query_success(self, e2e_dataset, tmpdir):
        # Create some files and an empty directory
        num_of_files = 25
        test_dir_name = "test_directory"
        empty_folder_name = "im_empty"
        full_folder_name = "im_full"

        os.chdir(tmpdir.strpath)
        os.mkdir(test_dir_name)
        os.mkdir(os.path.join(test_dir_name, empty_folder_name))
        os.mkdir(os.path.join(test_dir_name, full_folder_name))

        for i in range(num_of_files):
            filename = "file_{}".format(i)
            with open(os.path.join(test_dir_name, filename), "w") as f:
                f.write(str(i))

        for i in range(num_of_files):
            filename = "file_{}".format(i)
            with open(os.path.join(test_dir_name, full_folder_name, filename), "w") as f:
                f.write(str(i))

        files_to_upload = "{}/*".format(test_dir_name)
        e2e_dataset.put_files(paths=[files_to_upload])

        # Create query and clone
        query = TestQueries.create_query(e2e_dataset, "empty-query", "*{}/".format(empty_folder_name))
        e2e_dataset.clone(query_slug=query.slug)

        # Assertions
        cloned_test_dir_path = os.path.join(e2e_dataset.slug, test_dir_name)
        files = os.listdir(cloned_test_dir_path)
        count_folders = 0
        count_files = 0

        for file in files:
            if os.path.isdir(os.path.join(cloned_test_dir_path, file)):
                count_folders += 1
            else:
                count_files += 1

        assert count_folders == 1
        assert count_files == 0
        assert os.path.isdir(os.path.join(cloned_test_dir_path, empty_folder_name))

    def test_clone_files_using_folder_query_success(self, e2e_dataset, tmpdir):
        # Create some files and an empty directory
        num_of_files = 150
        test_dir_name = "test_directory"
        empty_folder_name = "im_empty"
        full_folder_name = "im_full"

        os.chdir(tmpdir.strpath)
        os.mkdir(test_dir_name)
        os.mkdir(os.path.join(test_dir_name, empty_folder_name))
        os.mkdir(os.path.join(test_dir_name, full_folder_name))

        for i in range(num_of_files):
            filename = "file_{}".format(i)
            with open(os.path.join(test_dir_name, filename), "w") as f:
                f.write(str(i))

        for i in range(num_of_files):
            filename = "file_{}".format(i)
            with open(os.path.join(test_dir_name, full_folder_name, filename), "w") as f:
                f.write(str(i))

        files_to_upload = "{}/*".format(test_dir_name)
        e2e_dataset.put_files(paths=[files_to_upload])

        # Create query and clone
        query = TestQueries.create_query(e2e_dataset, "full-query", "*{}*".format(full_folder_name))
        e2e_dataset.clone(query_slug=query.slug)

        # Assertions
        cloned_test_dir_path = os.path.join(e2e_dataset.slug, test_dir_name)

        assert os.path.isdir(os.path.join(cloned_test_dir_path, full_folder_name))
        assert not os.path.exists(os.path.join(cloned_test_dir_path, empty_folder_name))
        assert len(list(os.listdir(os.path.join(cloned_test_dir_path, full_folder_name)))) == num_of_files



import requests

from abc import ABC, abstractmethod
from cnvrgv2.utils.storage_utils import decrypt
from cnvrgv2.config import Config
from cnvrgv2.utils.converters import convert_bytes


class BaseStorageClient(ABC):
    def __init__(self, storage_meta):
        sts_url = storage_meta.get("sts_url")

        sts_content = requests.get(sts_url, verify=False).text
        sts_array = sts_content.split("\n")
        self.iv = sts_array[1]
        self.key = sts_array[0]
        self.check_certificate = Config().check_certificate or False

    def progress_callback(self, progress_bar):
        def progress_incrementer(bytes_received):
            if progress_bar and progress_bar.max > 0:
                converted_bytes, unit = convert_bytes(bytes_received, unit=progress_bar.unit)
                with progress_bar.mutex:
                    progress_bar.throttled_next(converted_bytes)

        return progress_incrementer

    def _decrypt_dict(self, props, keys=None):
        """
        Decrypt storage meta from the server.
        @param props: the storage meta dictionary
        @param keys: keys we want to decrypt
        @return: decrypted dictionary
        """

        # Only decrypts keys in props, otherwise leaves value unchanged
        return {k: decrypt(self.key, self.iv, v) if k in keys else v for k, v in props.items()}

    @abstractmethod
    def upload_single_file(self, local_path, object_path):
        """
        Uploads a single file using the relevant client
        @param local_path: The local path from which to upload
        @param object_path: The remote file path in the object storage service
        @return: None
        """
        pass

    @abstractmethod
    def download_single_file(self, local_path, object_path):
        """
        Downloads a single file using the relevant client
        @param local_path: The local path to which to download
        @param object_path: The remote file path in the object storage service
        @return:
        """
        pass



from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes


class Folder(DynamicAttributes):

    available_attributes = {
        "name": str,
        "fullpath": str,
        "created_at": str,
        "updated_at": str,
        "object_path": str
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._attributes = attributes or {}



import click
from prettytable import PrettyTable

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.config import error_messages
from cnvrgv2.errors import CnvrgHttpError
from cnvrgv2.utils.validators import validate_user_role


@click.group(name='members')
def members_group():
    pass


@members_group.command()
@click.option('-e', '--email', prompt=messages.MEMBER_ENTER_EMAIL, help=messages.MEMBER_HELP_EMAIL)
@click.option('-r', '--role', prompt=messages.MEMBER_ENTER_ROLE, help=messages.MEMBER_HELP_ROLE)
@prepare_command()
def add(cnvrg, logger, email, role):
    """
    Creates a new memberships
    """

    if not validate_user_role(role):
        logger.log_and_echo(error_messages.MEMBER_NOT_VALID_ROLE.format(role), error=True)

    cnvrg.members.add(email=email, role=role)
    success_message = messages.MEMBER_ADDED_SUCCESS.format(email, role)
    logger.log_and_echo(success_message)


@members_group.command(name='list')
@prepare_command()
def list_command(cnvrg):
    """
    list all members
    """
    members = cnvrg.members.list()
    table = PrettyTable()
    table.field_names = ["Username", "Email", "Role"]
    table.align["Username"] = "l"
    table.align["email"] = "l"
    for member in members:
        table.add_row([member.username, member.email, member.role.capitalize()])
    click.echo(table)


@members_group.command(name='revoke')
@click.option('-e', '--email', prompt=messages.MEMBER_ENTER_EMAIL, help=messages.MEMBER_HELP_EMAIL)
@prepare_command()
def revoke(cnvrg, logger, email):
    """
    removes a memberships
    """
    try:
        cnvrg.members.get(email).revoke()
    except CnvrgHttpError as error:
        logger.log_and_echo(error, error=True)
        return

    success_message = messages.MEMBER_REMOVED_SUCCESS.format(email)
    logger.log_and_echo(success_message)


@members_group.command(name="update")
@click.option('-e', '--email', prompt=messages.MEMBER_ENTER_EMAIL, help=messages.MEMBER_HELP_EMAIL)
@click.option('-r', '--role', prompt=messages.MEMBER_ENTER_ROLE, help=messages.MEMBER_HELP_ROLE)
@prepare_command()
def update(cnvrg, logger, email, role):
    """
    update a membership
    """

    if not validate_user_role(role):
        logger.log_and_echo(error_messages.MEMBER_NOT_VALID_ROLE.format(role), error=True)

    cnvrg.members.get(email=email).update(role=role)
    success_message = messages.MEMBER_UPDATED_SUCCESS.format(email, role)
    logger.log_and_echo(success_message)



# TODO: Root folder is a folder that cnvrg adds to each data owner.
#  A bug was found in the server function that lists folders.
#  The function is data_commit.rb:data_trees and probably also commit.rb:trees
#  Those functions return the root directory ONLY if state_saved is true.
#  Since Those parts should go through a refactor pretty soon (11.2022), it wasn't fixed,
#  and the root folder is added here to the filter list. Once those functions are no more in use,
#  please re-check if the root folder can be removed.

CNVRG_METADATA_PATHS = [".cnvrgignore", "/"]


def clean_cnvrg_metadata_files_and_folders(original_generator):
    """
    For test purpose ONLY!
    Removes files and folders created by cnvrg from the given generator
    @param original_generator: Generator containing files and folders to clean
    @return: A filtered generator
    """
    return filter(lambda asset: asset.fullpath not in CNVRG_METADATA_PATHS, original_generator)



from cnvrgv2.config import routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.modules.base.workflow_instance_base import WorkflowInstanceBase
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.json_api_format import JAF
from cnvrgv2.utils.url_utils import urljoin


class NotebookType:
    JUPYTER_LAB = "jupyterlab"
    VS_CODE = "vscode"
    R_STUDIO = "r_studio"


class Workspace(WorkflowInstanceBase):

    available_attributes = {
        "jupyter_token": str,
        "iframe_url": str,
        "is_spark": bool,
        "spark_url": str,
        "spark_master_url": str,
        "notebook_type": str,
        "tensorboard_url": str,
        "commits": list,
        "webapps": dict,
        "endpoints": list,
        "experiments": list,
        "current_step": str,
        "workspace_steps": str,
        "on_stop_build": dict,
        "full_href": str,
        **WorkflowInstanceBase.available_attributes
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = Context(context=context)

        # Set current context scope to current project
        if slug:
            self._context.set_scope(SCOPE.WORKSPACE, slug)

        scope = self._context.get_scope(SCOPE.WORKSPACE)

        self._proxy = Proxy(context=self._context)
        self._route = routes.WORKSPACE_BASE.format(scope["organization"], scope["project"], scope["notebooksession"])
        self._attributes = attributes or {}
        self._type = "NotebookSession"
        self.slug = scope["notebooksession"]

    def build_image(self, registry, repository, tag="latest", is_custom=False, link_to_workspace=True):
        """
        Build an image from the current workspace
        @param registry: Name of the registry
        @param repository: Name of the repository in the registry
        @param tag: A tag for the image
        @param is_custom:
        @param link_to_workspace: If true, workspace will start from the linked image
        @return: None
        """

        build_image_url = urljoin(self._route, routes.WORKSPACE_BUILD_IMAGE_SUFFIX)

        attributes = {
            "registry_slug": registry,
            "repo": repository,
            "tag": tag,
            "is_custom": is_custom,
            "link_to_workspace": link_to_workspace
        }

        self._proxy.call_api(
            route=build_image_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        )

    def activate_build_image_on_stop(self, registry, repository, tag="latest", link_to_workspace=True):
        """
        Build image when workspace is stopped
        @param registry: Name of the registry
        @param repository: Name of the repository in the registry
        @param tag: A tag for the image
        @param link_to_workspace: If true, workspace will start from the linked image
        @return: None
        """
        activate_build_image_url = urljoin(self._route, routes.WORKSPACE_ACTIVATE_BUILD_IMAGE_ON_STOP_SUFFIX)

        attributes = {
            "registry_slug": registry,
            "repo": repository,
            "tag": tag,
            "activate": True,
            "is_custom": True,
            "link_to_workspace": link_to_workspace,
        }

        response_attributes = self._proxy.call_api(
            route=activate_build_image_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes=attributes)
        ).attributes

        self._attributes = {**self._attributes, **response_attributes}

    def disable_build_image_on_stop(self):
        """
        Stop automatically build when workspace is stopped
        @return: None
        """
        disable_build_image_url = urljoin(self._route, routes.WORKSPACE_DISABLE_BUILD_IMAGE_ON_STOP_SUFFIX)

        response_attributes = self._proxy.call_api(
            route=disable_build_image_url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes={})
        ).attributes

        self._attributes = {**self._attributes, **response_attributes}

    def start_tensorboard(self):
        """
        Starts a tensorboard server in the current workspace
        """
        url = urljoin(self._route, routes.WORKflOW_START_TENSORBOARD_SUFFIX)

        self._proxy.call_api(
            route=url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes={})
        )

    def stop_tensorboard(self):
        """
        Stop the tensorboard session in the current workspace
        """
        url = urljoin(self._route, routes.WORKflOW_STOP_TENSORBOARD_SUFFIX)

        self._proxy.call_api(
            route=url,
            http_method=HTTP.POST,
            payload=JAF.serialize(type=self._type, attributes={})
        )

    def restart(self):
        """
        Override restart from workflows_base to remove functionality.
        restart() is only relevant for webapps
        """
        raise AttributeError("'Workspace' object has no attribute 'restart'")



import click

from prettytable import PrettyTable

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.errors import CnvrgError
from cnvrgv2.cli.utils.helpers import print_generator_in_chunks


@click.group(name='labels')
def labels_group():
    """
    Manage labels in cnvrg
    """
    pass


@labels_group.command()
@click.option('-k', '--kind', help=messages.LABEL_HELP_KIND)
@click.option('-n', '--name', help=messages.LABEL_HELP_NAME)
@click.option('-c', '--color_name', help=messages.LABEL_HELP_COLOR_NAME)
@prepare_command()
def create(cnvrg, logger, kind, name, color_name):
    try:
        label = cnvrg.labels.create(name, kind, color_name)
        logger.log_and_echo(messages.LABEL_CREATE_SUCCESS.format(label.name))
    except CnvrgError as error:
        logger.log_and_echo(message=str(error), error=True)


@labels_group.command()
@click.option('-k', '--kind', prompt=messages.LABEL_PROMPT_KIND, help=messages.LABEL_HELP_KIND)
@click.option('-cs', '--chunk-size', default=20, help=messages.DATASET_LIST_FILES_HELP_CHUNK_SIZE)
@click.option('-l', '--limit', type=int, help=messages.DATASET_LIST_FILES_HELP_LIMIT)
@prepare_command()
def list(cnvrg, kind, chunk_size, limit):
    """
    Lists Labels
    """
    labels = cnvrg.labels.list(kind)
    attributes_field_name = ["Name", "Color Name"]
    attributes = ["name", "color_name"]
    print_generator_in_chunks(labels, chunk_size, limit, attributes, attributes_field_name, line_numbers=True)


@labels_group.command()
@click.option('-k', '--kind', help=messages.LABEL_HELP_KIND)
@click.option('-n', '--name', help=messages.LABEL_HELP_NAME)
@prepare_command()
def get(cnvrg, logger, kind, name):
    """
    Retrieves an Label information
    """
    try:
        label = cnvrg.labels.get(name=name, kind=kind)
    except CnvrgError as error:
        logger.log_and_echo(message=str(error), error=True)
        return

    table = PrettyTable()
    table.field_names = ["Name", "Color Name"]
    table.add_row([label.name, label.color_name])
    click.echo(table)


@labels_group.command()
@click.option('-k', '--kind', help=messages.LABEL_HELP_KIND)
@click.option('-n', '--name', help=messages.LABEL_HELP_NAME)
@prepare_command()
def delete(cnvrg, logger, name, kind):
    """
    Delete an label
    """
    try:
        label = cnvrg.labels.get(name=name, kind=kind)
        label.delete()
        msg = messages.LABEL_DELETE_SUCCESS.format(name)
        logger.log_and_echo(msg)
    except CnvrgError as error:
        logger.log_and_echo(message=str(error), error=True)
        return



import os
import shutil

import pytest
import yaml

from cnvrgv2 import Project
from cnvrgv2.config import Config, CONFIG_FOLDER_NAME, error_messages
from cnvrgv2.config import CONFIG_FILE_NAME
from cnvrgv2.data.file_downloader import cache_file
from cnvrgv2.errors import CnvrgFileError
from tests.conftest import random_string
from tests.e2e.helpers.storage_helper import clean_cnvrg_metadata_files_and_folders


class DataOwnerFiles:

    def data_owner_verify_cnvrg_files_exists(self, data_owner, tmpdir):
        os.chdir(tmpdir.strpath)
        data_owner.clone()
        assert os.path.exists(os.path.join(data_owner.slug, ".cnvrgignore"))
        assert os.path.exists(os.path.join(data_owner.slug, ".cnvrg/cnvrg.config"))

    def data_owner_put_files(self, data_owner, temp_file):
        data_owner.put_files([temp_file.strpath])
        files = data_owner.list_files()
        assert (len(list(files))) == 1

    def data_owner_put_unchanged_files(self, data_owner, temp_file):
        initial_last_commit = data_owner.last_commit
        local_commit = data_owner.local_commit
        assert (local_commit is None)
        data_owner.put_files([temp_file.strpath])
        data_owner.reload()
        # Local commit should be null
        local_commit_after_first_put = data_owner.local_commit
        assert (local_commit_after_first_put is None)
        # Last commit should be updated to the new commit
        last_commit_after_first_put = data_owner.last_commit
        assert (initial_last_commit != last_commit_after_first_put)
        # Upload the same file once again without any changes
        # It should not create a commit (no changes have been made)
        data_owner.put_files([temp_file.strpath])
        data_owner.reload()
        # Last commit should be the same as the previous commit : no new commit should have been created
        last_commit_after_second_put = data_owner.last_commit
        assert (last_commit_after_second_put == last_commit_after_first_put)
        # Local commit should be null
        local_commit_after_second_put = data_owner.local_commit
        assert (local_commit_after_second_put is None)

    def data_owner_put_folder(self, data_owner, tmpdir):
        # Add one file to tmpdir
        tmpdir.join("temp.txt")
        data_owner.put_files([tmpdir.strpath])

        files = data_owner.list_files()
        for file in files:
            if tmpdir.strpath.lstrip("/") in file.fullpath:  # Server saves files without leading /
                return
        assert False

    def data_owner_put_files_override_false(self, data_owner, temp_file):
        data_owner.put_files([temp_file.strpath])
        files_after_first_upload = data_owner.list_files()
        first_file = next(files_after_first_upload)

        # Try to upload the same file again, override set to False (by default)
        data_owner.put_files([temp_file.strpath])
        files_after_second_upload = data_owner.list_files()
        second_file = next(files_after_second_upload)

        assert len(list(files_after_second_upload)) == len(list(files_after_first_upload))
        assert first_file.created_at == second_file.created_at

    def data_owner_put_files_override_true(self, data_owner, temp_file):
        data_owner.put_files([temp_file.strpath])
        files_after_first_upload = data_owner.list_files()
        first_file = next(files_after_first_upload)

        # Try to upload the same file again, override set to True
        data_owner.put_files([temp_file.strpath], override=True)
        files_after_second_upload = data_owner.list_files()
        second_file = next(files_after_second_upload)

        assert len(list(files_after_second_upload)) == len(list(files_after_first_upload))
        assert first_file.created_at != second_file.created_at

    def data_owner_put_files_abs_path_dir_path(self, data_owner, temp_file):
        # Add one file to tmpdir upload from absolute path that is not pwd
        dir_path = "destination1/destination2"
        data_owner.put_files(temp_file.strpath, dir_path=dir_path)
        files = list(data_owner.list_files())
        assert len(files) == 1
        assert files[0].fullpath.startswith(dir_path), \
            "File {0} does not start with the expected dir_path".format(temp_file.strpath)

    def data_owner_put_file_dir_path_relative_path(self, data_owner, temp_file):
        dir_path = "destination"
        os.chdir(os.path.dirname(temp_file))
        data_owner.put_files(paths=os.path.basename(temp_file), dir_path=dir_path)
        files = list(data_owner.list_files())
        assert len(files) == 1
        assert files[0].fullpath in os.path.join(dir_path, os.path.basename(temp_file)), \
            "File {0} not uploaded".format(files[0].fullpath)

    def data_owner_put_folder_abs_path_dir_path(self, data_owner, tmpdir):
        dir_path = "destination"
        data_owner.put_files([tmpdir.strpath], dir_path=dir_path)
        folders = list(data_owner.list_folders())
        assert len(folders) > 0
        for folder in folders:
            # Exclude root folder when checking the list of dataset folders.
            if folder.name != "root":
                assert folder.fullpath.startswith(dir_path), f"Folder {tmpdir.strpath} not uploaded"

    def data_owner_put_folder_dir_path_relative_path(self, data_owner, tmpdir):
        folder = "relative_folder_testing"
        dir_path = "destination"
        os.chdir(tmpdir)
        if not os.path.exists(folder):
            os.makedirs(folder)
        with open(os.path.join(folder, "file_from_folder2.txt"), 'w') as file:
            file.write("content: this file not empty")
        data_owner.put_files(paths=folder, dir_path=dir_path)
        folders = list(data_owner.list_folders())
        assert len(folders) > 0
        for folder in folders:
            # Exclude root folder when checking the list of dataset folders.
            if folder.name != "root":
                assert folder.fullpath.startswith(dir_path), \
                    "Folder {0} does not start with dir_path".format(tmpdir)
        files = list(data_owner.list_files())
        assert len(files) == 1

    def data_owner_put_empty_folder_dir_path_relative_path(self, data_owner, tmpdir):
        folder = "empty"
        dir_path = "destination"
        os.chdir(tmpdir)
        if not os.path.exists(folder):
            os.makedirs(folder)
        data_owner.put_files(paths=folder, dir_path=dir_path)
        folders = list(data_owner.list_folders())
        assert len(folders) > 0
        for folder in folders:
            # Exclude root folder when checking the list of dataset folders.
            if folder.name != "root":
                assert folder.fullpath.startswith(dir_path), \
                    "Folder {0} does not start with dir_path".format(tmpdir)
        files = list(data_owner.list_files())
        assert len(files) == 0

    def data_owner_put_files_force(self, data_owner, temp_file, temp_file_second):
        commit1 = data_owner.put_files([temp_file.strpath, temp_file_second.strpath])
        filenames = [file.file_name for file in data_owner.list_files()]
        assert temp_file.basename in filenames
        assert temp_file_second.basename in filenames
        assert len(filenames) == 2

        commit2 = data_owner.put_files([temp_file.strpath], force=True)
        filenames = [file.file_name for file in data_owner.list_files()]
        assert temp_file.basename in filenames
        assert temp_file_second.basename not in filenames
        assert len(filenames) == 1
        assert commit1 != commit2

    def data_owner_list_files(self, data_owner, tmpdir):
        num_files_to_upload = 5

        # create files and upload them
        files_to_upload = []
        for i in range(num_files_to_upload):
            file = tmpdir.join("temp{}.txt".format(i))
            file.write("content")
            files_to_upload.append(file.strpath)
        data_owner.put_files(files_to_upload)

        files = data_owner.list_files()
        assert (len(list(files))) == num_files_to_upload

    def data_owner_list_files_var(self, data_owner, tmpdir, num_of_files):
        files_to_upload = []
        filenames_to_upload = []
        for i in range(num_of_files):
            file = tmpdir.join("temp{}.txt".format(i))
            with open(file, "w") as f:
                f.write("file data{}".format(i))

            files_to_upload.append(file.strpath)
            filenames_to_upload.append(os.path.basename(file.strpath))

        data_owner.put_files(files_to_upload, force=True)
        files_retrieved = [file.file_name for file in data_owner.list_files()]

        assert len(filenames_to_upload) == num_of_files
        assert sorted(files_retrieved) == sorted(filenames_to_upload)

    def data_owner_list_files_and_folders(self, data_owner, tmpdir):
        num_files_and_folders = 5
        folders_to_upload = []

        os.chdir(tmpdir.strpath)

        for i in range(num_files_and_folders):
            test_folder_path = tmpdir.mkdir("test_folder_{}".format(i))
            test_file = test_folder_path.join("temp{}.txt".format(i))
            test_file.write("content")
            folders_to_upload.append(test_folder_path.basename)

        data_owner.put_files(folders_to_upload)
        files_and_folders = clean_cnvrg_metadata_files_and_folders(data_owner.list())

        assert (len(list(files_and_folders))) == num_files_and_folders * 2

    def data_owner_delete_files(self, data_owner, temp_file):
        data_owner.put_files([temp_file.strpath])

        data_owner.remove_files([temp_file.strpath])
        files = data_owner.list_files()
        assert (len(list(files))) == 0

    def data_owner_clone_files(self, data_owner, tmpdir):
        os.chdir(tmpdir.strpath)
        data_owner.clone()
        assert not data_owner._validate_config_ownership()
        os.chdir(data_owner.slug)
        data_owner.reload()
        assert data_owner._validate_config_ownership()
        assert os.path.exists(CONFIG_FOLDER_NAME)

    def data_owner_clone_with_empty_folder(self, data_owner, tmpdir):
        os.chdir(tmpdir.strpath)

        empty_folder_name = "empty"
        os.mkdir(empty_folder_name)
        data_owner.put_files(paths=[empty_folder_name])

        data_owner.clone()

        os.chdir(data_owner.slug)
        assert os.path.isdir(empty_folder_name)

    def data_owner_clone_override(self, data_owner, tmpdir, temp_file):
        # preparing 2 folders, with the same file name and diffrent content.
        os.chdir(tmpdir.strpath)
        os.mkdir("folder1")
        file = tmpdir.join("folder1/t1.txt")
        file.write("t1data")

        os.mkdir("folder2")
        file = tmpdir.join("folder2/t1.txt")
        file.write("t2data")

        # upload first file, and clone
        os.chdir("folder1")
        data_owner.put_files(paths=["t1.txt"])
        data_owner.clone()

        # on different folder, upload, and clone
        os.chdir("../folder2")
        data_owner.put_files(paths=["t1.txt"])
        os.chdir("../folder1/")
        data_owner.clone(override=True)

        os.chdir(data_owner.slug)
        assert open('t1.txt').read() == "t2data"

    def data_owner_upload_files(self, data_owner, tmpdir):
        os.chdir(tmpdir.strpath)

        new_file = 'new.txt'
        deleted_file = 'deleted.txt'

        # Init one file in the data owner and make sure it exists
        open(deleted_file, "a").close()
        data_owner.put_files([deleted_file])
        data_owner_files = list(data_owner.list_files())
        assert data_owner_files[0].fullpath == deleted_file

        # Clone the data_owner and enter the new folder
        data_owner.clone()
        os.chdir(data_owner.slug)

        # check number of files locally (including .cnvrg + .cnvrgignore) and remove the tmp file
        assert len(os.listdir()) == 3
        os.remove(deleted_file)
        assert len(os.listdir()) == 2

        # Add new file and upload it
        open(new_file, "a").close()
        data_owner.upload()

        # Now we should have only the new file in the data owner without the deleted file
        files_after_upload = list(clean_cnvrg_metadata_files_and_folders(data_owner.list_files()))
        assert len(files_after_upload) == 1
        assert files_after_upload[0].fullpath == new_file

    def data_owner_upload_unchanged_files(self, data_owner, tmpdir):
        os.chdir(tmpdir.strpath)

        new_file = 'new.txt'
        deleted_file = 'deleted.txt'

        # Init one file in the data owner and make sure it exists
        open(deleted_file, "a").close()
        data_owner.put_files([deleted_file])
        data_owner_files = list(data_owner.list_files())
        # Initial local commit should be null
        initial_local_commit = data_owner.local_commit
        assert (initial_local_commit is None)
        assert data_owner_files[0].fullpath == deleted_file

        # Clone the data_owner and enter the new folder
        data_owner.clone()
        os.chdir(data_owner.slug)

        # check number of files locally (including .cnvrg + .cnvrgignore) and remove the tmp file
        assert len(os.listdir()) == 3
        os.remove(deleted_file)
        assert len(os.listdir()) == 2

        # Add new file and upload it
        open(new_file, "a").close()
        data_owner.upload()

        # Now we should have only the new file in the data owner without the deleted file
        files_after_upload = list(data_owner.list_files())
        assert len(list(files_after_upload)) == 2  # With .cnvrgingore
        assert len([f for f in files_after_upload if f.fullpath == new_file]) == 1
        # Local commit should be updated to the new commit
        local_commit_after_first_upload = data_owner.local_commit
        assert (initial_local_commit != local_commit_after_first_upload)
        # Upload the same file once again without any changes
        # It should not create a commit (no changes have been made)
        data_owner.upload()
        local_commit_after_second_upload = data_owner.local_commit
        # Local commit should be the same as the previous commit : no new commit should have been created
        assert (local_commit_after_second_upload == local_commit_after_first_upload)

    def data_owner_upload_files_from_non_cnvrg(self, data_owner, tmpdir):
        os.chdir(tmpdir.strpath)
        tmpdir.join("temp.txt")

        with pytest.raises(CnvrgFileError) as exception_info:
            data_owner._config = Config()
            data_owner.upload()

        assert error_messages.CONFIG_YAML_NOT_FOUND == str(exception_info.value)

    def data_owner_download_files(self, data_owner):
        tmp_file = 'temp.txt'
        try:
            data_owner.clone()
            os.chdir(data_owner.slug)

            # Upload new file to project, then download
            open(tmp_file, "a").close()
            data_owner.put_files([tmp_file])
            os.remove(tmp_file)
            data_owner.download()
            files_after_download = [f.name for f in os.scandir(os.getcwd()) if f.name != CONFIG_FOLDER_NAME]

            assert len(files_after_download) == 2

        finally:
            # Cleanups
            if data_owner.slug in os.getcwd():
                os.chdir('..')
            if os.path.exists(data_owner.slug):
                shutil.rmtree(data_owner.slug)

    def data_owner_download_empty_folder(self, data_owner, tmpdir):
        os.chdir(tmpdir.strpath)
        empty_dir_name = "empty"
        os.mkdir(empty_dir_name)
        data_owner.clone()
        data_owner.put_files(paths=[empty_dir_name])

        os.chdir(data_owner.slug)
        data_owner.download()

        assert os.path.isdir(empty_dir_name)

    def data_owner_sync_local(self, data_owner):
        tmp_file1 = 'temp1.txt'
        tmp_file2 = 'temp2.txt'
        try:
            data_owner.clone()

            # Upload one file to project
            open(tmp_file1, 'a').close()
            data_owner.put_files([tmp_file1])

            # Add one file to local project dir
            os.chdir(data_owner.slug)
            open(os.getcwd() + "/" + tmp_file2, "a").close()

            commit_sha1 = data_owner.sync()
            remote_files_after_sync = [f.file_name for f in data_owner.list_files()]
            local_files_after_sync = [f.name for f in os.scandir(os.getcwd()) if f.name != CONFIG_FOLDER_NAME]

            assert all(f in remote_files_after_sync for f in local_files_after_sync)
            assert commit_sha1 == data_owner.local_commit
            assert commit_sha1 is not None

        finally:
            # Cleanups
            if data_owner.slug in os.getcwd():
                os.chdir('..')
            if os.path.exists(tmp_file1):
                os.remove(tmp_file1)
            if os.path.exists(data_owner.slug):
                shutil.rmtree(data_owner.slug)

    def data_owner_sync_local_with_output_dir(self, data_owner):
        tmp_file1 = 'temp1.txt'
        tmp_file2 = 'temp2.txt'
        tmp_file3 = 'temp3.txt'

        try:
            data_owner.clone()

            # Upload one file to project
            open(tmp_file1, 'a').close()
            data_owner.put_files([tmp_file1])

            # Add one file to local project dir
            os.chdir(data_owner.slug)
            open(os.getcwd() + "/" + tmp_file2, "a").close()

            # Add one file to local project's output dir
            os.mkdir('output')
            open(os.getcwd() + "/output/" + tmp_file3, "a").close()

            data_owner.sync(output_dir='output')
            remote_files_after_sync = [f.file_name for f in data_owner.list_files()]
            expected_files = [tmp_file1, tmp_file3]

            assert all(f in remote_files_after_sync for f in expected_files)
            assert not any(set(remote_files_after_sync) - set(expected_files))

        finally:
            # Cleanups
            if data_owner.slug in os.getcwd():
                os.chdir('..')
            if os.path.exists(tmp_file1):
                os.remove(tmp_file1)
            if os.path.exists(data_owner.slug):
                shutil.rmtree(data_owner.slug)

    def data_owner_put_files_absolute_path(self, tmpdir, data_owner):
        os.chdir(tmpdir.strpath)
        filename = random_string(5)
        try:
            open(filename, 'w').close()
            absolute_path = os.path.abspath(filename)
            data_owner.put_files([absolute_path], message="test project files fullpath")

            files = data_owner.list_files()
            for file in files:
                if file.fullpath == absolute_path.lstrip("/"):  # Server saves files without leading /
                    return

            assert False
        finally:
            if os.path.exists(filename):
                os.remove(filename)

    def data_owner_remove_files_absolute_path(self, tmpdir, data_owner):
        os.chdir(tmpdir.strpath)
        filename = random_string(5)
        try:
            open(filename, 'w').close()
            absolute_path = os.path.abspath(filename)
            data_owner.put_files([absolute_path], message="test dataset files fullpath")

            data_owner.remove_files([absolute_path])

            files = data_owner.list_files()
            assert len(list(files)) == 0
        finally:
            if os.path.exists(filename):
                os.remove(filename)

    def data_owner_remove_folders_through_upload(self, tmpdir, data_owner):
        os.chdir(tmpdir.strpath)
        folder_name = random_string(5)
        filename = random_string(5)

        os.mkdir(folder_name)
        open(folder_name + "/" + filename, 'w').close()
        data_owner.put_files([folder_name])

        # check file and folder were created on remote
        check_folder_exist = False
        for i in list(data_owner.list_folders()):
            if folder_name in i._attributes["fullpath"]:
                check_folder_exist = True

        check_file_exist = False
        for i in list(data_owner.list_files()):
            if filename in i._attributes["fullpath"]:
                check_file_exist = True

        assert check_folder_exist
        assert check_file_exist

        data_owner.clone()
        os.chdir(data_owner.slug)
        shutil.rmtree(folder_name)

        data_owner.upload()

        folders = clean_cnvrg_metadata_files_and_folders(data_owner.list_folders())
        files = clean_cnvrg_metadata_files_and_folders(data_owner.list_files())
        assert len(list(files)) == 0
        assert len(list(folders)) == 0

    def data_owner_remove_file_through_upload(self, tmpdir, data_owner):
        os.chdir(tmpdir.strpath)
        filename = random_string(5)

        open(filename, 'w').close()
        data_owner.put_files([filename])

        check_file_exist = False
        for i in list(data_owner.list_files()):
            if i._attributes["fullpath"] == filename:
                check_file_exist = True
        assert check_file_exist

        data_owner.clone()
        os.chdir(data_owner.slug)
        os.remove(filename)

        data_owner.upload()
        folders = clean_cnvrg_metadata_files_and_folders(data_owner.list_folders())
        assert len(list(folders)) == 0

    def data_owner_remove_folder_locally_through_download(self, tmpdir, data_owner):
        os.chdir(tmpdir.strpath)
        folder_name = random_string(5)
        filename = random_string(5)

        os.mkdir(folder_name)
        open(folder_name + "/" + filename, 'w').close()
        data_owner.put_files([folder_name])
        data_owner.clone()
        data_owner.remove_files([folder_name + "/"])

        os.chdir(data_owner.slug)
        data_owner.download()
        assert not os.path.isdir(folder_name)

    def data_owner_remove_file_locally_through_download(self, tmpdir, data_owner):
        os.chdir(tmpdir.strpath)
        filename = random_string(5)

        open(filename, 'w').close()
        data_owner.put_files([filename])
        data_owner.clone()
        data_owner.remove_files([filename])

        os.chdir(data_owner.slug)
        data_owner.download()
        assert not os.path.isfile(filename)

    def data_owner_sync_complex(self, tmpdir, data_owner):

        def create_full(paths):
            for path in paths:
                if "/" in path:
                    os.makedirs(os.path.dirname(path))

                if not path.endswith("/"):
                    open(path, "a").close()

        init_paths = ["a/del_server/c/f1", "del_folder_local/f2", "remove_on_server", "remove_local",
                      "mod_local_del_server", "modified_both"]
        os.chdir(tmpdir.strpath)
        create_full(init_paths)

        data_owner.put_files(["."])
        data_owner.clone()

        modified_both = open("modified_both", "a")
        modified_both.write("file changed remotely")
        modified_both.close()
        data_owner.put_files(["modified_both"])

        create_full(["new_folder_server/file.txt"])
        data_owner.put_files(["new_folder_server/"])

        os.chdir(data_owner.slug)

        os.remove("remove_local")
        shutil.rmtree("del_folder_local")

        mod_local_del_server = open("mod_local_del_server", "a")
        mod_local_del_server.write("file changed locally")
        mod_local_del_server.close()

        modified_both = open("modified_both", "a")
        modified_both.write("file changed locally")
        modified_both.close()

        data_owner.remove_files(["remove_on_server"])
        data_owner.remove_files(["mod_local_del_server"])
        data_owner.remove_files(["a/b"])

        create_full(["new_folder_local/file.txt"])

        commit_sha1 = data_owner.sync()

        # testing
        assert commit_sha1 == data_owner.local_commit
        assert commit_sha1 is not None

        server_folders = []
        for folder in list(data_owner.list_folders()):
            server_folders.append(folder._attributes["fullpath"])

        server_files = []
        for file in list(data_owner.list_files()):
            server_files.append(file._attributes["fullpath"])

        assert os.path.exists("modified_both.conflict")
        assert os.path.exists("new_folder_server")
        assert "new_folder_local/" in server_folders
        assert "remove_local" not in server_files
        assert "del_folder_local/f2" not in server_files
        assert "remove_on_server" not in server_folders

        with open('modified_both', 'r') as f:
            assert f.read() == "file changed locally"

        assert not os.path.exists("remove_on_server")
        assert not os.path.exists("mod_local_del_server")

    def data_owner_nested(self, tmpdir, data_owner):
        os.chdir(tmpdir.strpath)

        conf = yaml.dump({
            "commit_sha1": "test-commit",
            "git": True,
            "dataset_slug": "fake-slug",
        })

        os.makedirs('.cnvrg/')
        config_file = open('.cnvrg/' + CONFIG_FILE_NAME, 'a')

        config_file.write(conf)
        config_file.close()

        data_owner.clone()

        # check that local config is not corrupted
        config_file = Config()

        assert config_file.project_slug is None
        assert config_file.dataset_slug == "fake-slug"

        # check that local config is not corrupted
        os.chdir(data_owner.slug)
        config_file = Config()

        if isinstance(data_owner, Project):
            assert config_file.project_slug is not None
            assert config_file.dataset_slug is None
            assert config_file.project_slug == data_owner.slug
        else:
            assert config_file.dataset_slug is not None
            assert config_file.project_slug is None
            assert config_file.dataset_slug == data_owner.slug

        assert not data_owner._validate_config_ownership()

    def data_owner_cache(self, tmpdir, data_owner):
        os.chdir(tmpdir.strpath)

        def create_files(paths):
            for path in paths:
                if "/" in path:
                    os.makedirs(os.path.dirname(path), exist_ok=True)

                if not path.endswith("/"):
                    open(path, "a").close()

        init_paths = ["a/dataset_name/a.txt", "a/dataset_name/b.txt", "b/dataset_name/"]

        create_files(init_paths)

        os.chdir("b/dataset_name/")

        cached = cache_file("dataset_name", "a.txt", "a.txt", "a", back_path="../../") and os.path.exists("a.txt")

        not_cached = not cache_file("dataset_name", "c.txt", "c.txt", "a", back_path="../../") and not os.path.exists(
            "c.txt")

        assert cached and not_cached

    def data_owner_sync_removed_file(self, data_owner, tmpdir):
        os.chdir(tmpdir.strpath)
        filename = random_string(5)
        open(filename, "a")
        data_owner.put_files([filename])
        data_owner.clone()
        data_owner.remove_files([filename])

        os.chdir(data_owner.slug)
        commit_sha1 = data_owner.sync()

        assert commit_sha1 == data_owner.local_commit
        assert commit_sha1 is not None
        assert os.path.exists(filename) is False

    def data_owner_sync_removed_folder(self, data_owner, tmpdir):
        os.chdir(tmpdir.strpath)
        folder_name = random_string(5)
        os.mkdir(folder_name)
        filename = random_string(5)
        open("{}/{}".format(folder_name, filename), "a")
        data_owner.put_files([folder_name])
        data_owner.clone()
        data_owner.remove_files([folder_name + "/"])

        os.chdir(data_owner.slug)
        commit_sha1 = data_owner.sync()

        assert commit_sha1 == data_owner.local_commit
        assert commit_sha1 is not None
        assert os.path.exists(folder_name) is False

    def data_owner_zero_size_file(self, data_owner):
        old_working_dir = os.getcwd()
        try:
            # Clone Git Project
            data_owner.clone()
            os.chdir(data_owner.slug)

            # Create zero size file
            folder_name = random_string(5)
            os.mkdir(folder_name)
            filename = random_string(5)
            open("{}/{}".format(folder_name, filename), "a")

            # Upload zero size file
            data_owner.sync(progress_bar_enabled=True, output_dir=folder_name)

            # Check if file successfully uploaded
            files = data_owner.list_files()
            files_list = list(files)
            assert filename == files_list[0].file_name

        except Exception as e:
            assert False, "Error: {}".format(e)

        finally:
            os.chdir(old_working_dir)
            if os.path.isdir(data_owner.slug):
                shutil.rmtree(data_owner.slug)

    def data_owner_list_commits(self, data_owner, temp_file):
        inserted_commits = [data_owner.put_files(paths=[temp_file.strpath], message='commit message 1')]
        commits_generator = data_owner.list_commits()
        commits = []
        for commit in commits_generator:
            commits.append(commit.sha1)
        assert set(inserted_commits).issubset(commits)

    def data_owner_access_all_commit_fields(self, data_owner, temp_file, second_temp_file):
        """
        This test was added after a bug where exception was thrown when message was empty (the init commit)
        """
        data_owner.put_files(paths=[temp_file.strpath])  # empty commit message
        data_owner.put_files(paths=[second_temp_file.strpath], message="committing another file")

        for commit in data_owner.list_commits():
            attributes = list(commit.available_attributes.keys())
            for attribute in attributes:
                commit.__getattr__(attribute)



import os
import re

import yaml

from cnvrgv2.config import error_messages
from cnvrgv2.errors import CnvrgArgumentsError, CnvrgFileError
from cnvrgv2.modules.users.user import ROLES


def validate_url(url):
    regex = re.compile(
        r'^(?:http|ftp)s?://'  # http:// or https://
        r'(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|'  # domain...
        r'localhost|'  # localhost...
        r'\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})'  # ...or ip
        r'(?::\d+)?'  # optional port
        r'(?:/?|[/?]\S+)$', re.IGNORECASE)

    return re.match(regex, url) is not None


def validate_email(email):
    regex = r'^[\w.-]+[@]\w+(\.\w+)*\.\w{2,5}$'
    return re.search(regex, email) is not None


def validate_user_role(role):
    return role in ROLES.ALL_ROLES


def validate_secret_key(key):
    regex = r'^[a-zA-Z_][a-zA-Z0-9_]*$'
    return re.match(regex, key) is not None


def validate_username(username):
    regex = r'^[A-Za-z0-9_-]+$'
    return re.search(regex, username) is not None


def validate_directory_name(path):
    regex = r"^[a-zA-Z0-9_.()'!\/-]+$"
    is_dir = os.path.isdir(path)
    if is_dir and not re.match(regex, path):
        raise CnvrgArgumentsError(error_messages.NOT_A_VALID_DIRECTORY_NAME.format(path))


def validate_types_in_list(input_list, input_type):
    """
    Validates that the input is a list of objects from the given type
    @param input_list: The input (should be a list of input_type objects)
    @param input_type: The type of the objects in input_list
    @return: Whenever input_list is a list of input_types
    """
    if not isinstance(input_list, list):
        return False
    for obj in input_list:
        if not isinstance(obj, input_type):
            return False
    return True


def attributes_validator(
        available_attributes,
        attributes,
        required_values=None,
):
    """
    @param available_attributes: Attributes available to the object that is being validated
    @param attributes: Attributes to validate
    @param required_values: Attributes that must be presented in the action
    @return: void, will raise errors on failed validations
    """

    required_values = required_values or []

    # Validates all the required attributes were received (and are not None or equivalent)
    for key in required_values:
        if key not in attributes.keys() or not attributes[key]:
            raise CnvrgArgumentsError(error_messages.MISSING_REQUIRED_VALUE.format(key))

    for key, value in attributes.items():
        # Validates that only allowed attributes were received
        if key == 'slug' or key not in available_attributes.keys():
            raise CnvrgArgumentsError(error_messages.FAULTY_KEY.format(key))
        # Validate received attribute types
        if value and not isinstance(value, available_attributes[key]):
            raise CnvrgArgumentsError(error_messages.FAULTY_VALUE.format(key))


# COMPUTES
def validate_gpu(gpu_obj):
    AVAILABLE_MIG_TYPES = [
        None,
        'mig-1g.5g',
        'mig-2g.20g',
        'mig-3g.25g'
    ]

    gpu = gpu_obj["count"]
    mig_device = gpu_obj["mig_device"]
    if mig_device not in AVAILABLE_MIG_TYPES:
        raise CnvrgArgumentsError(error_messages.TEMPLATE_FAULTY_MIG_VALUE)
    if gpu and not isinstance(gpu, float):
        raise CnvrgArgumentsError(error_messages.TEMPLATE_FAULTY_GPU.format(gpu))
    if mig_device and not isinstance(mig_device, str):
        raise CnvrgArgumentsError(error_messages.TEMPLATE_FAULTY_MIG.format(mig_device))


def validate_template_type(template_type):
    AVAILABLE_TEMPLATE_TYPES = [
        'regular',
        'mpi',
        'kubernetes',
        'pytorch',
        'spark_on_kubernetes',
        'ray',
        'ray_modin'
    ]

    if template_type and template_type not in AVAILABLE_TEMPLATE_TYPES:
        raise CnvrgArgumentsError(error_messages.TEMPLATE_FAULTY_TYPE)


def validate_hyper_parameters(parameters):
    types = ["discrete", "float", "integer", "categorical"]
    if parameters and not isinstance(parameters, list):
        raise CnvrgArgumentsError(error_messages.EXPERIMENT_FAULTY_PARAMETERS)

    # Check if there is duplicates in the parameters key
    keys = [param.get("key") for param in parameters]
    if len(keys) != len(set(keys)):
        raise CnvrgArgumentsError(error_messages.EXPERIMENT_DUPLICATE_PARAMETERS.format(keys))

    for param in parameters:
        # Validate parameter name
        if not param.get("key"):
            raise CnvrgArgumentsError(error_messages.EXPERIMENT_FAULTY_PARAMETERS_KEY)
        # Validate parameter type
        if param.get("type") not in types:
            raise CnvrgArgumentsError(error_messages.EXPERIMENT_FAULTY_PARAMETERS_TYPE)
        # Validate parameter values
        if param.get("type") == "discrete":
            validate_discrete_params(param.get("values"))
        # Validate parameter step is positive
        if param.get("steps") and param.get("steps") < 0:
            raise CnvrgArgumentsError(error_messages.EXPERIMENT_FAULTY_PARAMETERS_STEPS)


def validate_discrete_params(values):
    if not values:
        raise CnvrgArgumentsError(error_messages.EXPERIMENT_FAULTY_PARAMETERS_VALUES)
    if not isinstance(values, list):
        raise CnvrgArgumentsError(error_messages.EXPERIMENT_FAULTY_PARAMETERS_LIST)
    for value in values:
        if not isinstance(value, int) and not isinstance(value, float):
            raise CnvrgArgumentsError(error_messages.EXPERIMENT_FAULTY_PARAMETERS_INT)


def has_potential_csv_injection(user_input):
    """
    Validates that the given input doesn't start with characters that might indicate a csv injection
    @param user_input: The input to validate
    @return: bool. Whenever the first character might indicate a csv injection
    """
    forbidden_start_characters = re.compile(r'^[=+\-@\s\t\n\r,;\'"]+')
    return bool(forbidden_start_characters.match(user_input))


def _validate_yaml_path(yaml_path):
    """
    Checks the correctness of a YAML file
    @param yaml_path: Path of the yaml file
    @return: True or False
    """
    if not os.path.exists(yaml_path):
        raise CnvrgFileError(error_messages.NO_FILE)
    try:
        with open(yaml_path, 'r') as f:
            yaml.safe_load(f)
    except Exception:
        raise CnvrgFileError(error_messages.FLOW_GET_FAULTY_YAML)


def _validate_yaml_string(yaml_string):
    """
    Checks the correctness of a YAML string
    @param yaml_string: string of the yaml
    @return: True or False
    """
    if not yaml_string:
        raise CnvrgArgumentsError(error_messages.EMPTY_ARGUMENT.format("yaml_string"))
    try:
        yaml.safe_load(yaml_string)
    except Exception:
        raise CnvrgArgumentsError(error_messages.FLOW_GET_FAULTY_YAML)



import inspect
from functools import wraps

import click

from cnvrgv2 import Cnvrg
from cnvrgv2.cli.logger.logger import CnvrgLogger
from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.validators import verify_login
from cnvrgv2.config import Config
from cnvrgv2.config import error_messages
from cnvrgv2.errors import CnvrgError
from cnvrgv2.modules.workflows.endpoint.endpoint import Endpoint
from cnvrgv2.modules.workflows.experiment.experiment import Experiment
from cnvrgv2.modules.workflows.webapp.webapp import Webapp
from cnvrgv2.modules.workflows.workspace.workspace import Workspace


def prepare_command():
    """
    This decorator logs a start log for a cli command, can validate if the user is logged in, and injects
    the cnvrg object and logger object to the cli command function, if the arguments are expected by the function
    @return: the decorator function
    """

    def decorator(func):
        @wraps(func)
        def inner(*args, **kwargs):
            verify_login()
            cnvrg = Cnvrg()
            config = Config()
            # In case the user logged in through the sdk, the config file might not exist. config.save will create it
            config.save()
            logger = CnvrgLogger(click)
            command_name = func.__name__

            log_message = messages.LOG_START_COMMAND.format(command_name, str(kwargs))
            logger.info(log_message)

            func_args = inspect.getfullargspec(func).args
            inject_args = dict()

            def _try_load_project():
                project_slug = config.data_owner_slug or kwargs.get("project_slug")
                return cnvrg.projects.get(project_slug)

            if "cnvrg" in func_args:
                inject_args["cnvrg"] = cnvrg
            if "logger" in func_args:
                inject_args["logger"] = logger
            if "project" in func_args:
                inject_args["project"] = _try_load_project()

            if "dataset" in func_args:
                dataset_slug = config.data_owner_slug
                if not dataset_slug:
                    logger.log_and_echo(error_messages.LOCAL_CONFIG_MISSING_DATA_OWNER_SLUG, error=True)
                inject_args["dataset"] = cnvrg.datasets.get(dataset_slug)

            if "experiment" in func_args:
                try:
                    inject_args["experiment"] = Experiment()
                except CnvrgError:
                    slug = kwargs.get("slug")
                    project = _try_load_project()
                    inject_args["experiment"] = project.experiments.get(slug)

            if "workspace" in func_args:
                try:
                    inject_args["workspace"] = Workspace()
                except CnvrgError:
                    slug = kwargs.get("slug")
                    project = _try_load_project()
                    inject_args["workspace"] = project.workspaces.get(slug)

            if "endpoint" in func_args:
                try:
                    inject_args["endpoint"] = Endpoint()
                except CnvrgError:
                    slug = kwargs.get("slug")
                    project = _try_load_project()
                    inject_args["endpoint"] = project.endpoints.get(slug)

            if "webapp" in func_args:
                try:
                    inject_args["webapp"] = Webapp()
                except CnvrgError:
                    slug = kwargs.get("slug")
                    project = _try_load_project()
                    inject_args["webapp"] = project.webapps.get(slug)

            n_kwargs = {**kwargs, **inject_args}
            func(*args, **n_kwargs)

        return inner

    return decorator


def prepare_command_without_context():
    """
    This decorator logs a start log for a cli command, can validate if the user is logged in, and injects
    the cnvrg object and logger object to the cli command function, if the arguments are expected by the function
    @return: the decorator function
    """

    def decorator(func):
        @wraps(func)
        def inner(*args, **kwargs):
            logger = CnvrgLogger(click)
            command_name = func.__name__

            log_message = messages.LOG_START_COMMAND.format(command_name, str(kwargs))
            logger.info(log_message)

            func_args = inspect.getfullargspec(func).args
            inject_args = dict()

            if "logger" in func_args:
                inject_args["logger"] = logger
            n_kwargs = {**kwargs, **inject_args}
            func(*args, **n_kwargs)

        return inner

    return decorator



from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes


class ProjectCommit(DynamicAttributes):
    available_attributes = {
        "sha1": str,
        "source": str,
        "message": str,
        "created_at": str
    }

    def __init__(self, context=None, slug=None, attributes=None):
        self._context = context
        self._slug = slug
        self._attributes = attributes or {}



import click

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command
from cnvrgv2.errors import CnvrgFileError


@click.group(name='flow')
def flow_group():
    pass


@flow_group.command()
@click.option('-y', '--yaml-path', default=None, help=messages.FLOW_YAML_PATH)
@prepare_command()
def create(project, logger, yaml_path):
    """
    Creates a flow from a predefined yaml file
    """
    try:
        flow = project.flows.create(yaml_path=yaml_path)
        success_message = messages.FLOW_CREATE_SUCCESS.format(flow.title)
        logger.log_and_echo(success_message)
    except CnvrgFileError as e:
        logger.log_and_echo(str(e), error=True)



from cnvrgv2.modules.base.dynamic_attributes import DynamicAttributes


class DatasourceOperationsInterface(DynamicAttributes):
    def list_objects(self, page_size: int):
        raise NotImplementedError

    def download_file(self, file_path: str, destination_path: str, calc_destination_path: bool) -> None:
        raise NotImplementedError



from cnvrgv2.modules.labels.dataowner_labels_client import DataownerLabelsClient
from cnvrgv2.modules.labels.utils import LabelKind
from cnvrgv2.context import SCOPE
from cnvrgv2.config import routes


class ProjectLabelsClient(DataownerLabelsClient):
    def __init__(self, project):
        super().__init__(project)
        self.kind = LabelKind.PROJECT
        scope = self._context.get_scope(SCOPE.PROJECT)
        self._route = routes.LABELS_DATAOWNER_BASE.format(scope["organization"], self.kind, project.slug)



import pytest

from cnvrgv2.errors import CnvrgError
from cnvrgv2.modules.registries import RegistryTypes
from tests.e2e.conftest import call_database


class TestRegistries:
    @staticmethod
    def cleanup(context_prefix):
        delete_user_command = "DELETE FROM registries WHERE slug LIKE'{}%'".format(context_prefix)
        call_database(delete_user_command)

    def test_create_registry(self, class_context, e2e_client):
        title = class_context.generate_name(5)
        url = "docker.io"

        registry = e2e_client.registries.create(title=title, url=url)
        assert registry.slug
        assert registry.url == url
        assert registry.title == title
        assert registry.registry_type == RegistryTypes.OTHER

    def test_get_registry(self, class_context, e2e_client):
        registry = e2e_client.registries.get(slug="cnvrg")
        assert registry.url == "docker.io/cnvrg"
        assert registry.title == "cnvrg"

    def test_get_registry_images(self, class_context, e2e_client):
        registry = e2e_client.registries.get(slug="cnvrg")
        for image in registry.images:
            assert image.tag
            assert image.name
            assert image.registry_slug == 'cnvrg'
            assert image.registry_url == 'docker.io/cnvrg'

    def test_list_registries(self, class_context, e2e_client):
        registries = e2e_client.registries.list()
        cnvrg_registry = [registry for registry in registries if registry.slug == "cnvrg"][0]
        assert cnvrg_registry.url == "docker.io/cnvrg"
        assert cnvrg_registry.title == "cnvrg"

    def test_update_registry(self, class_context, e2e_client):
        title = class_context.generate_name(5)
        url = "docker.io"

        registry = e2e_client.registries.create(title=title, url=url)
        assert registry.url == url
        assert registry.title == title

        new_url = "good-url.io"
        registry.update(url=new_url)
        registry.reload()
        assert registry.url == new_url
        assert registry.title == title

    def test_delete_registry(self, class_context, e2e_client):
        title = class_context.generate_name(5)
        url = "docker.io"

        registry = e2e_client.registries.create(title=title, url=url)
        assert registry.slug

        registry.delete()
        with pytest.raises(CnvrgError):
            e2e_client.registries.get(slug=registry.slug).url  # Need to call attribute to bypass lazy load

    def test_delete_registry_from_client(self, class_context, e2e_client):
        title = class_context.generate_name(5)
        url = "docker.io"

        registry = e2e_client.registries.create(title=title, url=url)
        assert registry.slug

        e2e_client.registries.delete(slug=registry.slug)
        with pytest.raises(CnvrgError):
            e2e_client.registries.get(slug=registry.slug).url  # Need to call attribute to bypass lazy load



import pytest
from tests.unit.modules.datasource.conftest import bucket_files_without_path1, bucket_files_without_path2, bucket_files


class TestDatasourceList:

    # Run the tests multiple times, with different params
    # List with different paths
    @pytest.mark.parametrize(
        "datasource_path, expected_result",
        [
            ('path1', bucket_files_without_path1),
            ('path2', bucket_files_without_path2),
            ('', bucket_files),
            ('path1/', bucket_files_without_path1),
            ('path2/', bucket_files_without_path2),
            ('/', bucket_files),

        ],
        ids=['1 - path1',
             '2 - path2',
             '3 - no path',
             '4 - path1/',
             '5 - path2/',
             '6 - /'
             ]
    )
    def test_list_objects(self, datasource, mock_datasource_client, mock_datasource_client_paginator,
                          datasource_path, expected_result):
        """
        Parametrized test for testing list_objects with different paths and expected results.
        """
        datasource.path = datasource_path

        result = list(datasource.list_objects(page_size=100))

        # Assert that the result matches the expected files for the provided path
        assert result == expected_result

        # Verify the correct paginator was created
        datasource.client.get_paginator.assert_called_once_with('list_objects_v2')
        datasource.client.get_paginator().paginate.assert_called_once_with(
            Bucket=datasource.bucket_name, Prefix=datasource.path,
            PaginationConfig={'PageSize': 100}
        )



import requests

from cnvrgv2.config.config import SSO_VERSION
from cnvrgv2.context import Context
from cnvrgv2.proxy import HTTP, Proxy


class MockResponse:
    def __init__(self, status, json):
        self.status_code = status
        self._json = json

    def json(self, *args, **kwargs):
        return self._json


class TestProxy:
    def test_call_api(self, mocker):
        def mock_requests_post(*args, **kwargs):
            assert args[0] == "http://localhost:3000/api/sample/url"
            assert kwargs["headers"]["Authorization"] == "CAPI sampletoken"
            assert kwargs["headers"]["test"] == "test"
            assert kwargs["json"]["test"] == "test"

            return MockResponse(status=200, json={})

        def mock_credentials_init(obj):
            obj.domain = "http://localhost:3000"
            obj.token = "sampletoken"
            obj.organization = "test"
            obj.sso_version = SSO_VERSION.CAPI

        mocker.patch.object(Context, "__init__", mock_credentials_init)
        mocker.patch.object(requests, "post", mock_requests_post)

        creds = Context()
        proxy = Proxy(creds)
        proxy.call_api("sample/url", HTTP.POST, payload={"test": "test"}, headers={"test": "test"})



import pytest

from cnvrgv2.cnvrg import Cnvrg
from cnvrgv2.errors import CnvrgHttpError
from cnvrgv2.modules.organization.organizations_client import OrganizationsClient
from cnvrgv2.modules.users import ROLES


class TestMembers:

    def test_revoke_last_member_in_organization_fail(self, e2e_user, domain, e2e_client):
        org_client = OrganizationsClient(domain=domain, token=e2e_client._context.token)
        org = org_client.get(e2e_client._context.organization)
        # run the test if there is only one member in the organization
        if len(org["memberships"]) == 1:
            email = org["memberships"][0]['email'] or org["memberships"][0]['invite_email']
            with pytest.raises(CnvrgHttpError):
                e2e_client.members.get(email).revoke()

    def test_add_membership_for_non_existent_user_success(self, e2e_user, domain, e2e_client):
        email = "nonexistent@user.com"
        e2e_client.members.add(email, ROLES.MANAGER)
        org_client = OrganizationsClient(domain=domain, token=e2e_client._context.token)
        org = org_client.get(e2e_client._context.organization)
        assert not any(user["email"] == email for user in org["memberships"])
        assert any(member["invite_email"] == email for member in org["memberships"])

    def test_add_membership_for_existent_member_fail(self, e2e_user, domain, e2e_client):
        email = "nonexistent2@user.com"
        e2e_client.members.add(email, ROLES.MANAGER)
        # add membership once again for the same user
        with pytest.raises(CnvrgHttpError):
            e2e_client.members.add(email, ROLES.MANAGER)

    def test_get_members(self, domain, e2e_client):
        org_client = OrganizationsClient(domain=domain, token=e2e_client._context.token)
        org = org_client.get(e2e_client._context.organization)
        members = org["memberships"]
        assert len(members) >= 1

    def test_add_member_unauthorized(self, domain, e2e_client, e2e_data_scientist_user):
        client_of_non_admin_user = Cnvrg(
            domain=domain,
            email=e2e_data_scientist_user["email"],
            password=e2e_data_scientist_user["password"],
            organization=e2e_client._context.organization
        )
        email = "nonexistent@user.com"
        with pytest.raises(CnvrgHttpError):
            client_of_non_admin_user.members.add(email, ROLES.MANAGER)

    def test_revoke_member_success(self, e2e_user, domain, e2e_client, e2e_data_scientist_user):
        org_client = OrganizationsClient(domain=domain, token=e2e_client._context.token)
        e2e_client.members.get(e2e_data_scientist_user["email"]).revoke()
        org = org_client.get(e2e_client._context.organization)
        assert not any(user["email"] == e2e_data_scientist_user["email"] for user in org["memberships"])

    def test_revoke_non_existent_member_fail(self, e2e_user, e2e_client):
        with pytest.raises(CnvrgHttpError):
            e2e_client.members.get("nonexistent3@user.com").revoke()



from cnvrgv2 import cnvrg
from cnvrgv2.config import error_messages, routes
from cnvrgv2.context import Context, SCOPE
from cnvrgv2.errors import CnvrgArgumentsError
from cnvrgv2.modules.dataset import Dataset
from cnvrgv2.proxy import HTTP, Proxy
from cnvrgv2.utils.api_list_generator import api_list_generator
from cnvrgv2.utils.json_api_format import JAF


class DatasetsClient:
    def __init__(self, organization):
        if type(organization) is not cnvrg.Cnvrg:
            raise CnvrgArgumentsError(error_messages.ARGUMENT_BAD_TYPE.format(cnvrg.Cnvrg, type(organization)))
        self._context = Context(context=organization._context)
        scope = self._context.get_scope(SCOPE.ORGANIZATION)

        self._proxy = Proxy(context=self._context)
        self._route = routes.DATASETS_BASE.format(scope["organization"])

    def list(self, filter=None, sort="-id"):
        """
        List all datasets in a specific organization
        @param sort: key to sort the list by (-key -> DESC | key -> ASC)
        @param filter: Filter json
        @raise: HttpError
        @return: Generator that yields dataset objects
        """
        # TODO: Implement with commit/query and working_dir
        return api_list_generator(
            context=self._context,
            route=self._route,
            object=Dataset,
            sort=sort,
            filter=filter
        )

    def search(self, slug):
        """
        List all datasets where the slug contains the given slug
        @param slug: Substring of the desired dataset's slug
        @raise: HttpError
        @return: Generator that yields dataset objects
        """
        filter_obj = {
            "operator": "AND",
            "conditions": [
                {
                    "key": "slug",
                    "operator": "like",
                    "value": slug
                }
            ]
        }

        return self.list(filter=filter_obj)

    def get(self, slug):
        """
        Retrieves a dataset by the given slug
        @param slug: The slug of the requested dataset
        @return: Dataset object
        """
        if not slug or not isinstance(slug, str):
            raise CnvrgArgumentsError(error_messages.DATASET_GET_FAULTY_SLUG)

        return Dataset(context=self._context, slug=slug)

    def create(self, name, category="general"):
        """
        Creates a new dataset with the given name
        @param name: Name of the new dataset
        @param category: Dataset category [general, images, audio, video, text, tabular]
        @return: The newly created dataset object
        """
        if not name or not isinstance(name, str):
            raise CnvrgArgumentsError(error_messages.DATASET_CREATE_FAULTY_NAME)

        if not category or not isinstance(category, str):
            raise CnvrgArgumentsError(error_messages.DATASET_CREATE_FAULTY_CATEGORY)

        # TODO: Move list to const
        if category not in ["general", "images", "audio", "video", "text", "tabular"]:
            raise CnvrgArgumentsError(error_messages.DATASET_CREATE_FAULTY_CATEGORY_VALUE)

        attributes = {
            "title": name,
            "category": category
        }
        response = self._proxy.call_api(
            route=self._route,
            http_method=HTTP.POST,
            payload=JAF.serialize(type="dataset", attributes=attributes)
        )

        slug = response.attributes['slug']
        return Dataset(context=self._context, slug=slug)



import pytest
from tests.unit.modules.datasource.conftest import file_path


class TestDatasourceRemoveFile:

    def test_remove_file_success(self, datasource, mock_datasource_client, mock_normalize_file_path):
        normalized_path = mock_normalize_file_path.return_value
        datasource.client.delete_objects.return_value = {'Deleted': [{'Key': normalized_path}]}

        datasource.remove_file(file_path)

        datasource._normalize_file_path.assert_called_once_with(file_path)
        datasource.client.delete_objects.assert_called_once_with(
            Bucket=datasource.bucket_name,
            Delete={'Objects': [{'Key': normalized_path}]}
        )

    def test_remove_file_with_error(self, datasource, mock_datasource_client, mock_normalize_file_path):
        normalized_path = mock_normalize_file_path.return_value

        datasource.client.delete_objects.return_value = {
            'Errors': [{'Key': normalized_path, 'Message': 'ERROR!!!'}]
        }

        # Call the remove_file method and expect an exception
        with pytest.raises(Exception):
            datasource.remove_file(file_path)



import click

from cnvrgv2.cli.utils import messages
from cnvrgv2.cli.utils.decorators import prepare_command


@click.group(name='library')
def library_group():
    pass


@library_group.command()
@click.option('-l', '--library', prompt=messages.LIBRARY_PROMPT_CLONE, help=messages.PROJECT_PROMPT_CLONE)
@click.option('-o', '--override', is_flag=True, default=False, help=messages.PROJECT_HELP_CLONE_OVERRIDE)
@click.option('-d', '--working-dir', default=None, help=messages.OUTPUT_DIR_LOCATION)
@prepare_command()
def clone(cnvrg, logger, library, working_dir, override):
    """
    Clones the given project to local folder
    """
    if "==" in library:
        name = library.split("==")[0]
        version = library.split("==")[1]
    else:
        name = library
        version = "latest"

    library = cnvrg.libraries.get(name)
    library_version = library.versions.get(version)

    click.echo(messages.LOG_CLONING_LIBRARY.format(name, version))
    library_version.clone(progress_bar_enabled=True, working_dir=working_dir)

    success_message = messages.LIBRARY_CLONE_SUCCESS.format(name)
    logger.log_and_echo(success_message)
